# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import pytest
import scipp as sc
import scipp.constants
import scipp.testing
from numpy import pi

from scippneutron.chopper import DiskChopper, DiskChopperType


def deg_angle_to_time_factor(frequency: sc.Variable) -> sc.Variable:
    # Multiply by the returned value to convert an angle in degrees
    # to the time when the point at the angle reaches tdc.
    to_rad = sc.constants.pi / sc.scalar(180.0, unit='deg')
    angular_frequency = 2 * sc.constants.pi * abs(frequency.to(unit='Hz'))
    return sc.to_unit(to_rad / angular_frequency, 's/deg')


@pytest.fixture
def nexus_chopper():
    return sc.DataGroup(
        {
            'type': DiskChopperType.single,
            'position': sc.vector([0.0, 0.0, 2.0], unit='m'),
            'rotation_speed': sc.scalar(12.0, unit='Hz'),
            'beam_position': sc.scalar(45.0, unit='deg'),
            'phase': sc.scalar(-20.0, unit='deg'),
            'slit_edges': sc.array(
                dims=['slit'],
                values=[0.0, 60.0, 124.0, 126.0],
                unit='deg',
            ),
            'slit_height': sc.scalar(0.4, unit='m'),
            'radius': sc.scalar(0.5, unit='m'),
        }
    )


@pytest.mark.parametrize(
    'typ',
    [
        'contra_rotating_pair',
        'synchro_pair',
        DiskChopperType.contra_rotating_pair,
        DiskChopperType.synchro_pair,
    ],
)
def test_chopper_supports_only_single(nexus_chopper, typ):
    nexus_chopper['type'] = typ
    with pytest.raises(NotImplementedError):
        DiskChopper.from_nexus(nexus_chopper)


def test_frequency_must_be_frequency(nexus_chopper):
    nexus_chopper['rotation_speed'] = sc.scalar(1.0, unit='m/s')
    with pytest.raises(sc.UnitError):
        DiskChopper.from_nexus(nexus_chopper)


def test_eq(nexus_chopper):
    ch1 = DiskChopper.from_nexus(nexus_chopper)
    ch2 = DiskChopper.from_nexus(nexus_chopper)
    assert ch1 == ch2


@pytest.mark.parametrize(
    'replacement',
    [
        ('rotation_speed', sc.scalar(13.0, unit='Hz')),
        ('position', sc.vector([1, 0, 0], unit='m')),
        ('radius', sc.scalar(1.0, unit='m')),
        ('phase', sc.scalar(15, unit='deg')),
        ('slit_height', sc.scalar(0.14, unit='cm')),
        ('slit_edges', sc.array(dims=['slit'], values=[0.1, 0.3], unit='rad')),
    ],
)
def test_neq(nexus_chopper, replacement):
    ch1 = DiskChopper.from_nexus(nexus_chopper)
    ch2 = DiskChopper.from_nexus({**nexus_chopper, replacement[0]: replacement[1]})
    assert ch1 != ch2


def test_slit_begin_end_no_slit(nexus_chopper):
    ch = DiskChopper.from_nexus(
        {
            **nexus_chopper,
            'slit_edges': sc.zeros(sizes={'slit': 0}, unit='deg'),
        }
    )
    assert sc.identical(ch.slit_begin, sc.array(dims=['slit'], values=[], unit='deg'))
    assert sc.identical(ch.slit_end, sc.array(dims=['slit'], values=[], unit='deg'))


def test_slit_begin_end_one_slit(nexus_chopper):
    ch = DiskChopper.from_nexus(
        {
            **nexus_chopper,
            'slit_edges': sc.array(dims=['slit'], values=[13, 43], unit='deg'),
        }
    )
    assert sc.identical(ch.slit_begin, sc.array(dims=['slit'], values=[13], unit='deg'))
    assert sc.identical(ch.slit_end, sc.array(dims=['slit'], values=[43], unit='deg'))


def test_slit_begin_end_two_slits(nexus_chopper):
    ch = DiskChopper.from_nexus(
        {
            **nexus_chopper,
            'slit_edges': sc.array(dims=['slit'], values=[0, 60, 124, 126], unit='deg'),
        }
    )
    assert sc.identical(
        ch.slit_begin, sc.array(dims=['slit'], values=[0, 124], unit='deg')
    )
    assert sc.identical(
        ch.slit_end, sc.array(dims=['slit'], values=[60, 126], unit='deg')
    )


@pytest.mark.parametrize('rotation_speed', [1.0, -1.0])
def test_slit_begin_end_two_slits_unordered(nexus_chopper, rotation_speed):
    ch = DiskChopper.from_nexus(
        {
            **nexus_chopper,
            'slit_edges': sc.array(
                dims=['slit'], values=[2.5, 2.8, 0.8, 1.3], unit='rad'
            ),
        }
    )
    assert sc.identical(
        ch.slit_begin, sc.array(dims=['slit'], values=[2.5, 0.8], unit='rad')
    )
    assert sc.identical(
        ch.slit_end, sc.array(dims=['slit'], values=[2.8, 1.3], unit='rad')
    )


def test_slit_begin_end_across_0(nexus_chopper):
    ch = DiskChopper.from_nexus(
        {
            **nexus_chopper,
            'slit_edges': sc.array(dims=['slit'], values=[340.0, 382.0], unit='deg'),
        }
    )
    assert sc.identical(
        ch.slit_begin, sc.array(dims=['slit'], values=[340.0], unit='deg')
    )
    assert sc.identical(
        ch.slit_end, sc.array(dims=['slit'], values=[382.0], unit='deg')
    )


@pytest.mark.parametrize(
    'edges', [[0, 100, 60, 140], [30, 100, 0, 40], [0, 100, 100, 140]]
)
def test_overlapping_slits_from_combined_edges(nexus_chopper, edges):
    with pytest.raises(ValueError, match='overlap'):
        DiskChopper.from_nexus(
            {
                **nexus_chopper,
                'slit_edges': sc.array(dims=['slit'], values=edges, unit='deg'),
            }
        )


def test_overlapping_slits_from_separate_edges(nexus_chopper):
    del nexus_chopper['slit_edges']
    with pytest.raises(ValueError, match='overlap'):
        DiskChopper.from_nexus(
            {
                **nexus_chopper,
                'slit_begin': sc.array(dims=['slit'], values=[10, 80], unit='deg'),
                'slit_end': sc.array(dims=['slit'], values=[100, 150], unit='deg'),
            }
        )


@pytest.mark.parametrize(
    'edges', [[30, 40, 10, 100], [10, 100, 10, 20], [10, 100, 80, 100]]
)
def test_slits_contained_in_other_slit_from_combined_edges(nexus_chopper, edges):
    with pytest.raises(ValueError, match='overlap'):
        DiskChopper.from_nexus(
            {
                **nexus_chopper,
                'slit_edges': sc.array(dims=['slit'], values=edges, unit='deg'),
            }
        )


def test_slits_contained_in_other_slit_from_separate_edges(nexus_chopper):
    del nexus_chopper['slit_edges']
    with pytest.raises(ValueError, match='overlap'):
        DiskChopper.from_nexus(
            {
                **nexus_chopper,
                'slit_begin': sc.array(dims=['slit'], values=[10, 20], unit='deg'),
                'slit_end': sc.array(dims=['slit'], values=[100, 30], unit='deg'),
            }
        )


@pytest.mark.parametrize('beam_position_unit', ['rad', 'deg'])
@pytest.mark.parametrize('phase_unit', ['rad', 'deg'])
def test_time_offset_angle_at_beam_no_phase_zero_beam_pos_clockwise_single_angle(
    nexus_chopper, beam_position_unit, phase_unit
):
    ch = DiskChopper.from_nexus(
        {
            **nexus_chopper,
            'beam_position': sc.scalar(0.0, unit=beam_position_unit),
            'phase': sc.scalar(0.0, unit=phase_unit),
            'rotation_speed': sc.scalar(-2.3, unit='Hz'),
        }
    )
    omega = 2 * pi * 2.3
    # Note that the chopper performs one rotation prior to the first pulse, hence
    # the additional -2 * pi.
    expected = sc.array(dims=['slit'], values=[-2 * pi, 0.0], unit='s') / omega
    assert sc.allclose(
        ch.time_offset_angle_at_beam(angle=sc.scalar(0.0, unit='rad')), expected
    )
    assert sc.allclose(
        ch.time_offset_angle_at_beam(angle=sc.scalar(0.0, unit='deg')), expected
    )

    assert sc.allclose(
        ch.time_offset_angle_at_beam(angle=sc.scalar(1.2, unit='rad')),
        sc.array(dims=['slit'], values=[1.2 - 2 * pi, 1.2], unit='s') / omega,
    )
    assert sc.allclose(
        ch.time_offset_angle_at_beam(angle=sc.scalar(-0.4, unit='rad')),
        sc.array(dims=['slit'], values=[-0.4 - 2 * pi, -0.4], unit='s') / omega,
    )
    assert sc.allclose(
        ch.time_offset_angle_at_beam(angle=sc.scalar(7.1, unit='rad')),
        sc.array(dims=['slit'], values=[7.1 - 2 * pi, 7.1], unit='s') / omega,
    )


@pytest.mark.parametrize('beam_position_unit', ['rad', 'deg'])
@pytest.mark.parametrize('phase_unit', ['rad', 'deg'])
def test_time_offset_angle_at_beam_no_phase_zero_beam_pos_anti_clockwise_single_angle(
    nexus_chopper, beam_position_unit, phase_unit
):
    ch = DiskChopper.from_nexus(
        {
            **nexus_chopper,
            'beam_position': sc.scalar(0.0, unit=beam_position_unit),
            'phase': sc.scalar(0.0, unit=phase_unit),
            'rotation_speed': sc.scalar(2.3, unit='Hz'),
        }
    )
    omega = 2 * pi * 2.3
    # Note that the chopper performs one rotation prior to the first pulse, hence
    # the additional 0.0.
    expected = sc.array(dims=['slit'], values=[0.0, 2 * pi], unit='s') / omega
    assert sc.allclose(
        ch.time_offset_angle_at_beam(angle=sc.scalar(0.0, unit='rad')), expected
    )
    assert sc.allclose(
        ch.time_offset_angle_at_beam(angle=sc.scalar(0.0, unit='deg')), expected
    )
    assert sc.allclose(
        ch.time_offset_angle_at_beam(angle=sc.scalar(1.2, unit='rad')),
        sc.array(dims=['slit'], values=[-1.2, 2 * pi - 1.2], unit='s') / omega,
    )
    assert sc.allclose(
        ch.time_offset_angle_at_beam(angle=sc.scalar(-0.4, unit='rad')),
        sc.array(dims=['slit'], values=[0.4, 2 * pi + 0.4], unit='s') / omega,
    )
    assert sc.allclose(
        ch.time_offset_angle_at_beam(angle=sc.scalar(7.1, unit='rad')),
        sc.array(dims=['slit'], values=[-7.1, 2 * pi - 7.1], unit='s') / omega,
    )


@pytest.mark.parametrize('beam_position_unit', ['rad', 'deg'])
@pytest.mark.parametrize('phase_unit', ['rad', 'deg'])
def test_time_offset_angle_at_beam_no_phase_zero_beam_pos_clockwise_multi_angle(
    nexus_chopper, beam_position_unit, phase_unit
):
    ch = DiskChopper.from_nexus(
        {
            **nexus_chopper,
            'beam_position': sc.scalar(0.0, unit=beam_position_unit),
            'phase': sc.scalar(0.0, unit=phase_unit),
            'rotation_speed': sc.scalar(-4.4, unit='Hz'),
        }
    )
    omega = 2 * pi * 4.4
    angles = sc.array(dims=['angle'], values=[0.0, 0.4, -1.3, 6.9], unit='rad')
    # Note that the chopper performs one rotation prior to the first pulse, hence
    # the additional -2 * pi.
    # Note that order here is important for the flatten to happen in the correct order.
    expected = (
        sc.array(dims=['dummy'], values=[-2.0 * pi, 0.0], unit='s')
        + sc.array(dims=['angle'], values=[0.0, 0.4, -1.3, 6.9], unit='s')
    ).flatten(to='angle') / omega
    assert sc.allclose(ch.time_offset_angle_at_beam(angle=angles), expected)


@pytest.mark.parametrize('beam_position_unit', ['rad', 'deg'])
@pytest.mark.parametrize('phase_unit', ['rad', 'deg'])
def test_time_offset_angle_at_beam_no_phase_zero_beam_pos_anti_clockwise_multi_angle(
    nexus_chopper, beam_position_unit, phase_unit
):
    ch = DiskChopper.from_nexus(
        {
            **nexus_chopper,
            'beam_position': sc.scalar(0.0, unit=beam_position_unit),
            'phase': sc.scalar(0.0, unit=phase_unit),
            'rotation_speed': sc.scalar(4.4, unit='Hz'),
        }
    )
    omega = 2 * pi * 4.4
    # Note that the chopper performs one rotation prior to the first pulse, hence
    # the additional 0.0.
    offset = sc.array(dims=['dummy'], values=[0.0, 2.0 * pi], unit='s')
    angles = sc.array(dims=['angle'], values=[0.0, 0.4, -1.3, 6.9], unit='rad')
    expected = (
        offset - sc.array(dims=['angle'], values=[0.0, 0.4, -1.3, 6.9], unit='s')
    ).flatten(to='angle') / omega
    assert sc.allclose(ch.time_offset_angle_at_beam(angle=angles), expected)


@pytest.mark.parametrize('phase_unit', ['rad', 'deg'])
def test_time_offset_angle_at_beam_no_phase_with_beam_pos_clockwise(
    nexus_chopper, phase_unit
):
    ch = DiskChopper.from_nexus(
        {
            **nexus_chopper,
            'beam_position': sc.scalar(1.8, unit='rad'),
            'phase': sc.scalar(0.0, unit=phase_unit),
            'rotation_speed': sc.scalar(-2.3, unit='Hz'),
        }
    )
    omega = 2 * pi * 2.3
    assert sc.allclose(
        ch.time_offset_angle_at_beam(angle=sc.scalar(0.0, unit='rad')),
        sc.array(dims=['slit'], values=[-2.0 * pi - 1.8, -1.8], unit='s') / omega,
    )
    assert sc.allclose(
        ch.time_offset_angle_at_beam(angle=sc.scalar(1.8, unit='rad')),
        sc.array(dims=['slit'], values=[-2.0 * pi, 0.0], unit='s') / omega,
    )
    assert sc.allclose(
        ch.time_offset_angle_at_beam(angle=sc.scalar(2.4, unit='rad')),
        sc.array(dims=['slit'], values=[-2.0 * pi + 2.4 - 1.8, 2.4 - 1.8], unit='s')
        / omega,
    )
    assert sc.allclose(
        ch.time_offset_angle_at_beam(angle=sc.scalar(7.1, unit='rad')),
        sc.array(dims=['slit'], values=[-2.0 * pi + 7.1 - 1.8, 7.1 - 1.8], unit='s')
        / omega,
    )


@pytest.mark.parametrize('phase_unit', ['rad', 'deg'])
def test_time_offset_angle_at_beam_no_phase_with_beam_pos_anti_clockwise(
    nexus_chopper, phase_unit
):
    ch = DiskChopper.from_nexus(
        {
            **nexus_chopper,
            'beam_position': sc.scalar(1.8, unit='rad'),
            'phase': sc.scalar(0.0, unit=phase_unit),
            'rotation_speed': sc.scalar(2.3, unit='Hz'),
        }
    )
    omega = 2 * pi * 2.3
    assert sc.allclose(
        ch.time_offset_angle_at_beam(angle=sc.scalar(0.0, unit='rad')),
        sc.array(dims=['slit'], values=[1.8, 2 * pi + 1.8], unit='s') / omega,
    )
    assert sc.allclose(
        ch.time_offset_angle_at_beam(angle=sc.scalar(1.8, unit='rad')),
        sc.array(dims=['slit'], values=[0.0, 2 * pi], unit='s') / omega,
    )
    assert sc.allclose(
        ch.time_offset_angle_at_beam(angle=sc.scalar(2.4, unit='rad')),
        sc.array(dims=['slit'], values=[-2.4 + 1.8, 2 * pi - 2.4 + 1.8], unit='s')
        / omega,
    )
    assert sc.allclose(
        ch.time_offset_angle_at_beam(angle=sc.scalar(7.1, unit='rad')),
        sc.array(dims=['slit'], values=[-7.1 + 1.8, 2 * pi - 7.1 + 1.8], unit='s')
        / omega,
    )


@pytest.mark.parametrize('beam_position_unit', ['rad', 'deg'])
def test_time_offset_angle_at_beam_with_phase_zero_beam_pos_clockwise(
    nexus_chopper, beam_position_unit
):
    ch = DiskChopper.from_nexus(
        {
            **nexus_chopper,
            'beam_position': sc.scalar(0.0, unit=beam_position_unit),
            'phase': sc.scalar(-0.7, unit='rad'),
            'rotation_speed': sc.scalar(-1.1, unit='Hz'),
        }
    )
    omega = 2 * pi * 1.1
    assert sc.allclose(
        ch.time_offset_angle_at_beam(angle=sc.scalar(0.0, unit='rad')),
        sc.array(dims=['slit'], values=[-2.0 * pi + 0.7, 0.7], unit='s') / omega,
    )
    assert sc.allclose(
        ch.time_offset_angle_at_beam(angle=sc.scalar(1.8, unit='rad')),
        sc.array(dims=['slit'], values=[-2.0 * pi + 1.8 + 0.7, 1.8 + 0.7], unit='s')
        / omega,
    )
    assert sc.allclose(
        ch.time_offset_angle_at_beam(angle=sc.scalar(2.4, unit='rad')),
        sc.array(dims=['slit'], values=[-2.0 * pi + 2.4 + 0.7, 2.4 + 0.7], unit='s')
        / omega,
    )
    assert sc.allclose(
        ch.time_offset_angle_at_beam(angle=sc.scalar(7.1, unit='rad')),
        sc.array(dims=['slit'], values=[-2.0 * pi + 7.1 + 0.7, 7.1 + 0.7], unit='s')
        / omega,
    )


@pytest.mark.parametrize('beam_position_unit', ['rad', 'deg'])
def test_time_offset_angle_at_beam_with_phase_zero_beam_pos_anti_clockwise(
    nexus_chopper, beam_position_unit
):
    ch = DiskChopper.from_nexus(
        {
            **nexus_chopper,
            'beam_position': sc.scalar(0.0, unit=beam_position_unit),
            'phase': sc.scalar(-0.7, unit='rad'),
            'rotation_speed': sc.scalar(1.1, unit='Hz'),
        }
    )
    omega = 2 * pi * 1.1
    assert sc.allclose(
        ch.time_offset_angle_at_beam(angle=sc.scalar(0.0, unit='rad')),
        sc.array(dims=['slit'], values=[-0.7, 2.0 * pi - 0.7], unit='s') / omega,
    )
    assert sc.allclose(
        ch.time_offset_angle_at_beam(angle=sc.scalar(1.8, unit='rad')),
        sc.array(dims=['slit'], values=[-1.8 - 0.7, 2.0 * pi - 1.8 - 0.7], unit='s')
        / omega,
    )
    assert sc.allclose(
        ch.time_offset_angle_at_beam(angle=sc.scalar(2.4, unit='rad')),
        sc.array(dims=['slit'], values=[-2.4 - 0.7, 2.0 * pi - 2.4 - 0.7], unit='s')
        / omega,
    )
    assert sc.allclose(
        ch.time_offset_angle_at_beam(angle=sc.scalar(7.1, unit='rad')),
        sc.array(dims=['slit'], values=[-7.1 - 0.7, 2.0 * pi - 7.1 - 0.7], unit='s')
        / omega,
    )


def test_time_offset_open_close_no_slit(nexus_chopper):
    ch = DiskChopper.from_nexus(
        {
            **nexus_chopper,
            'phase': sc.scalar(-0.7, unit='rad'),
            'slit_edges': sc.zeros(
                sizes={'slit': 0},
                unit='rad',
            ),
        }
    )
    sc.testing.assert_identical(
        ch.time_offset_open(pulse_frequency=ch.frequency),
        sc.zeros(sizes={'slit': 0}, unit='s'),
    )
    sc.testing.assert_identical(
        ch.time_offset_close(pulse_frequency=ch.frequency),
        sc.zeros(sizes={'slit': 0}, unit='s'),
    )
    sc.testing.assert_identical(
        ch.open_duration(pulse_frequency=ch.frequency),
        sc.zeros(sizes={'slit': 0}, unit='s'),
    )


@pytest.mark.parametrize('rotation_speed', [5.12, -3.6])
def test_time_offset_open_close_only_slit(nexus_chopper, rotation_speed):
    ch = DiskChopper.from_nexus(
        {
            **nexus_chopper,
            'beam_position': sc.scalar(0.0, unit='rad'),
            'phase': sc.scalar(0.0, unit='rad'),
            'rotation_speed': sc.scalar(rotation_speed, unit='Hz'),
            'slit_edges': sc.array(
                dims=['slit'],
                values=[0.0, 360.0],
                unit='deg',
            ),
        }
    )
    factor = deg_angle_to_time_factor(sc.scalar(rotation_speed, unit='Hz'))
    assert sc.allclose(
        ch.time_offset_open(pulse_frequency=abs(ch.frequency)),
        sc.array(dims=['slit'], values=[-360.0, 0.0], unit='deg') * factor,
    )
    assert sc.allclose(
        ch.time_offset_close(pulse_frequency=abs(ch.frequency)),
        sc.array(dims=['slit'], values=[0.0, 360.0], unit='deg') * factor,
    )
    assert sc.allclose(
        ch.open_duration(pulse_frequency=abs(ch.frequency)),
        sc.array(dims=['slit'], values=[1 / abs(rotation_speed)] * 2, unit='s'),
    )


@pytest.mark.parametrize(
    'phase',
    [
        sc.scalar(0.0, unit='rad'),
        sc.scalar(1.2, unit='rad'),
        sc.scalar(-50.0, unit='deg'),
        sc.scalar(-50, unit='deg'),  # also check for integer phase
    ],
)
@pytest.mark.parametrize(
    'beam_position',
    [
        sc.scalar(0.0, unit='rad'),
        sc.scalar(-0.5, unit='rad'),
        sc.scalar(45.8, unit='deg'),
    ],
)
def test_time_offset_open_close_one_slit_clockwise(nexus_chopper, phase, beam_position):
    ch = DiskChopper.from_nexus(
        {
            **nexus_chopper,
            'beam_position': beam_position,
            'phase': phase,
            'rotation_speed': sc.scalar(-7.21, unit='Hz'),
            'slit_edges': sc.array(dims=['slit'], values=[87.0, 177.0], unit='deg'),
        }
    )
    factor = deg_angle_to_time_factor(sc.scalar(-7.21, unit='Hz'))
    shift = phase.to(unit='deg', dtype='float64') + beam_position.to(unit='deg')
    assert sc.allclose(
        ch.time_offset_open(pulse_frequency=abs(ch.frequency)),
        (sc.array(dims=['slit'], values=[87.0 - 360.0, 87.0], unit='deg') - shift)
        * factor,
    )
    assert sc.allclose(
        ch.time_offset_close(pulse_frequency=abs(ch.frequency)),
        (sc.array(dims=['slit'], values=[177.0 - 360.0, 177.0], unit='deg') - shift)
        * factor,
    )
    assert sc.allclose(
        ch.open_duration(pulse_frequency=abs(ch.frequency)),
        sc.array(dims=['slit'], values=[90.0] * 2, unit='deg') * factor,
    )


@pytest.mark.parametrize(
    'phase',
    [
        sc.scalar(0.0, unit='rad'),
        sc.scalar(1.2, unit='rad'),
        sc.scalar(-50.0, unit='deg'),
    ],
)
@pytest.mark.parametrize(
    'beam_position',
    [
        sc.scalar(0.0, unit='rad'),
        sc.scalar(-0.5, unit='rad'),
        sc.scalar(45.8, unit='deg'),
    ],
)
def test_time_offset_open_close_one_slit_anticlockwise(
    nexus_chopper, phase, beam_position
):
    ch = DiskChopper.from_nexus(
        {
            **nexus_chopper,
            'beam_position': beam_position,
            'phase': phase,
            'rotation_speed': sc.scalar(7.21, unit='Hz'),
            'slit_edges': sc.array(dims=['slit'], values=[87.0, 177.0], unit='deg'),
        }
    )
    factor = deg_angle_to_time_factor(sc.scalar(7.21, unit='Hz'))
    shift = phase.to(unit='deg') + beam_position.to(unit='deg')
    assert sc.allclose(
        ch.time_offset_open(pulse_frequency=ch.frequency),
        (sc.array(dims=['slit'], values=[-177.0, 360 - 177.0], unit='deg') + shift)
        * factor,
    )
    assert sc.allclose(
        ch.time_offset_close(pulse_frequency=ch.frequency),
        (sc.array(dims=['slit'], values=[-87.0, 360 - 87.0], unit='deg') + shift)
        * factor,
    )
    assert sc.allclose(
        ch.open_duration(pulse_frequency=ch.frequency),
        sc.array(dims=['slit'], values=[90.0] * 2, unit='deg') * factor,
    )


@pytest.mark.parametrize(
    'phase',
    [
        sc.scalar(0.0, unit='rad'),
        sc.scalar(1.2, unit='rad'),
        sc.scalar(-50.0, unit='deg'),
    ],
)
@pytest.mark.parametrize(
    'beam_position',
    [
        sc.scalar(0.0, unit='rad'),
        sc.scalar(-0.5, unit='rad'),
        sc.scalar(45.8, unit='deg'),
    ],
)
def test_time_offset_open_close_one_slit_across_tdc_clockwise(
    nexus_chopper, phase, beam_position
):
    ch = DiskChopper.from_nexus(
        {
            **nexus_chopper,
            'beam_position': beam_position,
            'phase': phase,
            'rotation_speed': sc.scalar(-7.21, unit='Hz'),
            'slit_edges': sc.array(dims=['slit'], values=[330.0, 380.0], unit='deg'),
        }
    )
    factor = deg_angle_to_time_factor(sc.scalar(-7.21, unit='Hz'))
    shift = phase.to(unit='deg') + beam_position.to(unit='deg')
    assert sc.allclose(
        ch.time_offset_open(pulse_frequency=abs(ch.frequency)),
        (sc.array(dims=['slit'], values=[330.0 - 360.0, 330.0], unit='deg') - shift)
        * factor,
    )
    assert sc.allclose(
        ch.time_offset_close(pulse_frequency=abs(ch.frequency)),
        (sc.array(dims=['slit'], values=[380.0 - 360.0, 380.0], unit='deg') - shift)
        * factor,
    )
    assert sc.allclose(
        ch.open_duration(pulse_frequency=abs(ch.frequency)),
        sc.array(dims=['slit'], values=[50.0] * 2, unit='deg') * factor,
    )


@pytest.mark.parametrize(
    'phase',
    [
        sc.scalar(0.0, unit='rad'),
        sc.scalar(1.2, unit='rad'),
        sc.scalar(-50.0, unit='deg'),
    ],
)
@pytest.mark.parametrize(
    'beam_position',
    [
        sc.scalar(0.0, unit='rad'),
        sc.scalar(-0.5, unit='rad'),
        sc.scalar(45.8, unit='deg'),
    ],
)
def test_time_offset_open_close_one_slit_across_tdc_anticlockwise(
    nexus_chopper, phase, beam_position
):
    ch = DiskChopper.from_nexus(
        {
            **nexus_chopper,
            'beam_position': beam_position,
            'phase': phase,
            'rotation_speed': sc.scalar(7.21, unit='Hz'),
            'slit_edges': sc.array(dims=['slit'], values=[330.0, 380.0], unit='deg'),
        }
    )
    factor = deg_angle_to_time_factor(sc.scalar(7.21, unit='Hz'))
    shift = phase.to(unit='deg') + beam_position.to(unit='deg')
    assert sc.allclose(
        ch.time_offset_open(pulse_frequency=ch.frequency),
        (sc.array(dims=['slit'], values=[-380.0, 360 - 380.0], unit='deg') + shift)
        * factor,
    )
    assert sc.allclose(
        ch.time_offset_close(pulse_frequency=ch.frequency),
        (sc.array(dims=['slit'], values=[-330.0, 360 - 330.0], unit='deg') + shift)
        * factor,
    )
    assert sc.allclose(
        ch.open_duration(pulse_frequency=ch.frequency),
        sc.array(dims=['slit'], values=[50.0] * 2, unit='deg') * factor,
    )


def test_time_offset_open_close_two_slits_clockwise(nexus_chopper):
    ch = DiskChopper.from_nexus(
        {
            **nexus_chopper,
            'beam_position': sc.scalar(-32.0, unit='deg'),
            'phase': sc.scalar(0.0, unit='deg'),
            'rotation_speed': sc.scalar(-11.2, unit='Hz'),
            'slit_edges': sc.array(
                dims=['slit'],
                values=[87.0, 177.0, 280.0, 342.0],
                unit='deg',
            ),
        }
    )
    factor = deg_angle_to_time_factor(sc.scalar(-11.2, unit='Hz'))
    assert sc.allclose(
        ch.time_offset_open(pulse_frequency=abs(ch.frequency)),
        sc.array(
            dims=['slit'],
            values=[87.0 + 32 - 360.0, 280 + 32 - 360.0, 87.0 + 32, 280 + 32],
            unit='deg',
        )
        * factor,
    )
    assert sc.allclose(
        ch.time_offset_close(pulse_frequency=abs(ch.frequency)),
        sc.array(
            dims=['slit'],
            values=[177.0 + 32 - 360.0, 342 + 32 - 360.0, 177.0 + 32, 342 + 32],
            unit='deg',
        )
        * factor,
    )
    assert sc.allclose(
        ch.open_duration(pulse_frequency=abs(ch.frequency)),
        sc.array(dims=['slit'], values=[90.0, 62.0] * 2, unit='deg') * factor,
    )


def test_time_offset_open_close_two_slits_anticlockwise(nexus_chopper):
    ch = DiskChopper.from_nexus(
        {
            **nexus_chopper,
            'beam_position': sc.scalar(-32.0, unit='deg'),
            'phase': sc.scalar(0.0, unit='deg'),
            'rotation_speed': sc.scalar(11.2, unit='Hz'),
            'slit_edges': sc.array(
                dims=['slit'],
                values=[87.0, 177.0, 280.0, 342.0],
                unit='deg',
            ),
        }
    )
    factor = deg_angle_to_time_factor(sc.scalar(11.2, unit='Hz'))
    assert sc.allclose(
        ch.time_offset_open(pulse_frequency=ch.frequency),
        sc.array(
            dims=['slit'],
            values=[-177.0 - 32, -342 - 32, 360 - 177.0 - 32, 360 - 342 - 32],
            unit='deg',
        )
        * factor,
    )
    assert sc.allclose(
        ch.time_offset_close(pulse_frequency=ch.frequency),
        sc.array(
            dims=['slit'],
            values=[-87.0 - 32, -280 - 32, 360 - 87.0 - 32, 360 - 280 - 32],
            unit='deg',
        )
        * factor,
    )
    assert sc.allclose(
        ch.open_duration(pulse_frequency=ch.frequency),
        sc.array(dims=['slit'], values=[90.0, 62.0] * 2, unit='deg') * factor,
    )


def test_time_offset_open_close_two_slits_clockwise_two_pulses(nexus_chopper):
    ch = DiskChopper.from_nexus(
        {
            **nexus_chopper,
            'beam_position': sc.scalar(0.0, unit='deg'),
            'phase': sc.scalar(60.0, unit='deg'),
            'rotation_speed': sc.scalar(-14.0, unit='Hz'),
            'slit_edges': sc.array(
                dims=['slit'],
                values=[87.0, 177.0, 200.0, 240.0],
                unit='deg',
            ),
        }
    )
    pulse_frequency = sc.scalar(7.0, unit='Hz')
    factor = deg_angle_to_time_factor(sc.scalar(-14.0, unit='Hz'))
    assert sc.allclose(
        ch.time_offset_open(pulse_frequency=pulse_frequency),
        sc.array(
            dims=['slit'],
            values=[
                87.0 - 60 - 360,
                200.0 - 60 - 360,
                87.0 - 60,
                200.0 - 60,
                87.0 - 60 + 360,
                200.0 - 60 + 360,
            ],
            unit='deg',
        )
        * factor,
    )
    assert sc.allclose(
        ch.time_offset_close(pulse_frequency=pulse_frequency),
        sc.array(
            dims=['slit'],
            values=[
                177.0 - 60 - 360,
                240.0 - 60 - 360,
                177.0 - 60,
                240.0 - 60,
                177.0 - 60 + 360,
                240.0 - 60 + 360,
            ],
            unit='deg',
        )
        * factor,
    )
    assert sc.allclose(
        ch.open_duration(pulse_frequency=pulse_frequency),
        sc.array(dims=['slit'], values=[90.0, 40.0, 90.0, 40.0, 90.0, 40.0], unit='deg')
        * factor,
    )


def test_time_offset_open_close_two_slits_anticlockwise_two_pulses(nexus_chopper):
    ch = DiskChopper.from_nexus(
        {
            **nexus_chopper,
            'beam_position': sc.scalar(0.0, unit='deg'),
            'phase': sc.scalar(60.0, unit='deg'),
            'rotation_speed': sc.scalar(14.0, unit='Hz'),
            'slit_edges': sc.array(
                dims=['slit'],
                values=[87.0, 177.0, 200.0, 240.0],
                unit='deg',
            ),
        }
    )
    pulse_frequency = sc.scalar(7.0, unit='Hz')
    factor = deg_angle_to_time_factor(sc.scalar(14.0, unit='Hz'))
    assert sc.allclose(
        ch.time_offset_open(pulse_frequency=pulse_frequency),
        sc.array(
            dims=['slit'],
            values=[
                -177.0 + 60,
                -240.0 + 60,
                360 - 177.0 + 60,
                360 - 240.0 + 60,
                360 - 177.0 + 60 + 360,
                360 - 240.0 + 60 + 360,
            ],
            unit='deg',
        )
        * factor,
    )
    assert sc.allclose(
        ch.time_offset_close(pulse_frequency=pulse_frequency),
        sc.array(
            dims=['slit'],
            values=[
                -87.0 + 60,
                -200.0 + 60,
                360 - 87.0 + 60,
                360 - 200.0 + 60,
                360 - 87.0 + 60 + 360,
                360 - 200.0 + 60 + 360,
            ],
            unit='deg',
        )
        * factor,
    )
    assert sc.allclose(
        ch.open_duration(pulse_frequency=pulse_frequency),
        sc.array(dims=['slit'], values=[90.0, 40.0, 90.0, 40.0, 90.0, 40.0], unit='deg')
        * factor,
    )


def test_time_offset_open_close_two_slits_clockwise_half_pulse(nexus_chopper):
    ch = DiskChopper.from_nexus(
        {
            **nexus_chopper,
            'beam_position': sc.scalar(0.0, unit='deg'),
            'phase': sc.scalar(60.0, unit='deg'),
            'rotation_speed': sc.scalar(-3.5, unit='Hz'),
            'slit_edges': sc.array(
                dims=['slit'],
                values=[87.0, 177.0, 200.0, 240.0],
                unit='deg',
            ),
        }
    )
    pulse_frequency = sc.scalar(7.0, unit='Hz')
    factor = deg_angle_to_time_factor(sc.scalar(-3.5, unit='Hz'))
    assert sc.allclose(
        ch.time_offset_open(pulse_frequency=pulse_frequency),
        sc.array(
            dims=['slit'],
            values=[87.0 - 60 - 360, 200.0 - 60 - 360, 87.0 - 60, 200.0 - 60],
            unit='deg',
        )
        * factor,
    )
    assert sc.allclose(
        ch.time_offset_close(pulse_frequency=pulse_frequency),
        sc.array(
            dims=['slit'],
            values=[177.0 - 60 - 360, 240.0 - 60 - 360, 177.0 - 60, 240.0 - 60],
            unit='deg',
        )
        * factor,
    )
    assert sc.allclose(
        ch.open_duration(pulse_frequency=pulse_frequency),
        sc.array(dims=['slit'], values=[90.0, 40.0, 90.0, 40.0], unit='deg') * factor,
    )


def test_time_offset_open_close_source_frequency_not_multiple_of_chopper(nexus_chopper):
    ch = DiskChopper.from_nexus(
        {**nexus_chopper, 'rotation_speed': sc.scalar(4.52, unit='Hz')}
    )
    with pytest.raises(ValueError, match='out of phase'):
        ch.time_offset_open(pulse_frequency=sc.scalar(4.3, unit='Hz'))
    with pytest.raises(ValueError, match='out of phase'):
        ch.time_offset_close(pulse_frequency=sc.scalar(5.1, unit='Hz'))


def test_disk_chopper_svg(nexus_chopper):
    ch = DiskChopper.from_nexus(nexus_chopper)
    assert ch.make_svg()


def test_disk_chopper_svg_custom_dim_names(nexus_chopper):
    nexus_chopper['slit_edges'] = nexus_chopper['slit_edges'].rename_dims(slit='dim_0')
    ch = DiskChopper.from_nexus(nexus_chopper)
    assert ch.make_svg()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
from __future__ import annotations

import enum
import itertools
from collections.abc import Iterable
from dataclasses import dataclass
from typing import Literal

import numpy as np
import scipp as sc
from scipp.scipy.optimize import curve_fit
from scipy.stats import chi2 as _scipy_chi2

from ._common import FitParameters, FitRequirements
from .model import (
    GaussianModel,
    LorentzianModel,
    Model,
    PolynomialModel,
    PseudoVoigtModel,
)


@dataclass(eq=False, kw_only=True, slots=True)
class FitResult:
    """Optimized parameters and fit statistics for a single peak."""

    aic: sc.Variable
    """Akaike Information Criterion.

    A relative estimate of fit quality.
    Defined as

    .. math::

        \\mathsf{AIC} = 2k - 2\\ln(L)

    Where :math:`k` is the number of parameters and :math:`L` the likelihood
    if the model with the optimized parameters.
    """
    assessment: FitAssessment
    """Indicates whether the fit was successful or how if failed."""
    background: Model
    """Model for the background."""
    message: str
    """Short message describing the fit assessment."""
    p_value: sc.Variable
    """Probability of the given chi-squared or higher.

    The :math:`p`-value is the probability to get the same :math:`\\chi^2`
    or higher when repeating the fit.
    It is defined as

    .. math::

        p = 1 - F(\\chi^2;\\,\\nu)

    where :math:`F(\\chi^2;\\,\\nu)` is the cumulative distribution function
    of the :math:`\\chi^2`--distribution with :math:`\\nu` degrees of freedom.
    """
    peak: Model
    """Model for the peak."""
    popt: dict[str, sc.Variable]
    """Optimized parameters."""
    red_chisq: sc.Variable
    """Reduced chi-squared for the model and optimized parameters.

    .. math::

        \\chi^2_{\\nu} = \\frac1{\\nu} \\sum_{i=0}^n\\,
                         \\frac{(y_i - f(x_i))^2}{\\sigma_i^2}

    where :math:`\\nu` is the number of degrees of freedom.
    """
    window: sc.Variable
    """Fit window for this peak."""

    @classmethod
    def for_too_narrow_window(
        cls,
        *,
        peak: Model,
        background: Model,
        window: sc.Variable,
    ) -> FitResult:
        """Create a ``FitResult`` for a fit where the window is too narrow."""
        return cls.for_failure(
            assessment=FitAssessment.window_too_narrow,
            peak=peak,
            background=background,
            window=window,
        )

    @classmethod
    def for_failure(
        cls,
        *,
        assessment: FitAssessment | None = None,
        peak: Model,
        background: Model,
        window: sc.Variable,
        message: str | None = None,
    ) -> FitResult:
        """Create a ``FitResult`` for a failed fit."""
        return cls(
            popt={
                name: sc.scalar(np.nan)
                for name in peak.param_names | background.param_names
            },
            red_chisq=sc.scalar(-np.nan),
            p_value=sc.scalar(-np.nan),
            aic=sc.scalar(-np.inf),
            assessment=FitAssessment.failed if assessment is None else assessment,
            peak=peak,
            background=background,
            window=window,
            message=message or _message_from_assessment(assessment),
        )

    @property
    def success(self) -> bool:
        """Return whether the fit was successful."""
        return self.assessment == FitAssessment.success

    def better_than(self, other: FitResult) -> bool:
        """Return True if this result is better than another.

        Uses :attr:`aic` to compare the results.

        Parameters
        ----------
        other:
            Another ``FitResult`` to compare to.

        Returns
        -------
        :
            Whether this result is better than the other.
        """
        return sc.all(self.aic < other.aic).value

    def eval_model(self, x: sc.Variable) -> sc.Variable:
        """Evaluate the model with optimized parameters.

        Parameters
        ----------
        x:
            Independent variable.

        Returns
        -------
            The model evaluated at ``x`` with optimized parameters.
        """
        return (self.background + self.peak)(
            x, **{name: sc.values(val) for name, val in self.popt.items()}
        )

    def eval_peak(self, x: sc.Variable) -> sc.Variable:
        """Evaluate the peak model with optimized parameters.

        Parameters
        ----------
        x:
            Independent variable.

        Returns
        -------
            The peak model evaluated at ``x`` with optimized parameters.
        """
        return self.peak(
            x,
            **{
                name: sc.values(val)
                for name, val in self.popt.items()
                if name.startswith('peak_')
            },
        )

    def report(self) -> str:
        """Format an overview of the fit result."""
        param_name_length = max(map(len, self.popt.keys()))
        params = '\n- '.join(
            f'{{:{param_name_length}s}}: {val:c}'.format(key)
            for key, val in self.popt.items()
        )
        return f'''FitResult: {self.message}
In window: ({self.window[0].value}, {self.window[1].value}) {self.window.unit}
Models:
- background: {self.background.__class__.__name__}(prefix='{self.background.prefix}')
- peak: {self.peak.__class__.__name__}(prefix='{self.peak.prefix}')
Parameters:
- {params}
Statistics:
- AIC:              {self.aic.value:.2e}
- Red. chi-squared: {self.red_chisq.value:.2e}
- p:                {self.p_value.value:.3f}'''


class FitAssessment(enum.Enum):
    """Indicates whether the fit was successful or how if failed."""

    success = enum.auto()
    """The fit was successful."""
    failed = enum.auto()
    """The fit failed for some reason."""
    background_is_better = enum.auto()
    """The background fit yielded a better result than the background+peak fit."""
    peak_too_narrow = enum.auto()
    """The peak is too narrow given the resolution of the data."""
    peak_too_wide = enum.auto()
    """The peak is too wide given the size of the fit window."""
    peak_near_edge = enum.auto()
    """The peak is too close to the edge of the window."""
    peak_points_down = enum.auto()
    """The peak amplitude is negative."""
    p_too_small = enum.auto()
    """The p-value of the fit is below threshold."""
    window_too_narrow = enum.auto()
    """The fit window does not contain enough points."""


_BackgroundModelName = Literal['linear', 'quadratic']
_PeakModelName = Literal['gaussian', 'lorentzian', 'pseudo_voigt']


def fit_peaks(
    data: sc.DataArray,
    *,
    peak_estimates: sc.Variable,
    windows: sc.Variable,
    background: Model
    | _BackgroundModelName
    | Iterable[Model]
    | Iterable[_BackgroundModelName],
    peak: Model | _PeakModelName | Iterable[Model] | Iterable[_PeakModelName],
    fit_parameters: FitParameters | None = None,
    fit_requirements: FitRequirements | None = None,
) -> list[FitResult]:
    """Fit peaks to data.

    This function fits peaks to 1-dimensional data.
    The number of peaks and their approximate locations are determined
    by ``peak_estimates``.

    When defining fit windows, it is important that every
    window contains only one peak.
    Otherwise, fits may fail or converge to the wrong peak.
    Windows are, however, allowed to overlap as long as that
    overlap does not contain the bulk of a peak.

    Consider adjusting ``windows`` and ``fit_parameters`` if fits fail
    with all models.
    Only change ``fit_requirements`` if really needed as this may lead
    to bad fits being marked as succeeded.

    Fit procedure
    -------------
    For each peak, initial model parameters are estimated using
    :meth:`model.Model.guess` with the data within the fit window.
    Then, two fits are performed, one with only the background and one with
    background + peak using :func:`scipp.scipy.optimize.curve_fit`.

    If the latter fit converges, it is assessed based on several criteria like
    :attr:`FitRequirements.min_p_value` and the width of the obtained peak.
    In addition, the combined background + peak model must yield a better fit
    than the background alone (determined by their respective Akaike Information
    Criterion, see :attr:`FitResult.aic`).
    If the assessment fails, another set of models may be tried, see the section
    on model selection below.

    Finally, a :class:`FitResult` is constructed for each peak which contains the
    outcome of the assessment, the best parameters and models, and some goodness
    of fit statistics.
    Fits for each peak can succeed or fail independently.

    Model selection
    ---------------
    In the simplest case, the models for peaks and the background are given as a
    concrete model each, see :mod:`scippneutron.peaks.model`.
    Alternatively, they can be specified as strings.
    See the type annotations of the ``peak`` and ``background`` arguments for the lists
    of supported model names.

    If it is unclear which models are best in a given case or the best model varies
    between peaks, ``peak`` and ``background`` can be given as iterables of models
    or model names.
    In this case, the fit is attempted with each combination of peak and background
    model until a fit is successful.
    The background is varied first.

    For example, in

    .. code-block:: python

        fit_peaks(
            background=['linear', 'quadratic'],
            peak=['gaussian', 'lorentzian'],
            ...
        )

    models are tried in the order

    1. ``linear`` + ``gaussian``
    2. ``quadratic`` + ``gaussian``
    3. ``linear`` + ``lorentzian``
    4. ``quadratic`` + ``lorentzian``

    until the first successful fit.

    Parameters
    ----------
    data:
        A 1d data array where ``data.data`` is the dependent variable
        and ``data.coords[data.dim]`` is the independent variable for the fit.
        Must be 1-dimensional and not binned.
    peak_estimates:
        Initial estimates of peak locations.
        A peak will be fitted for each estimate.
        Must be a 1d variable with dimension ``data.dim``.
    windows:
        If a scalar, the size of fit windows.
        A window is constructed for each peak estimate centered on the estimate
        with a width equal to ``windows`` (adjusted to the data range and to maintain
        a separation between peaks, see
        :attr:`FitParameters.neighbor_separation_factor`).

        If a 2d array, the windows for each peak.
        Must have sizes ``{data.dim: len(data), 'range': 2}`` where
        ``windows['range', 0]`` and ``windows['range', 1]`` are the lower and upper
        bounds of the fit windows, respectively.
        The windows are not adjusted automatically in this case.
    background:
        The background model or models.
    peak:
        The peak model or models.
    fit_parameters:
        Parameters for the fit not otherwise listed as function arguments.
    fit_requirements:
        Constraints on the fit result.

    Returns
    -------
    :
        A :class:`FitResult` for each peak.
    """
    _assert_data_is_supported(data)

    background = _parse_model_spec(background, prefix='bkg_')
    peak = _parse_model_spec(peak, prefix='peak_')
    fit_parameters = fit_parameters or FitParameters()
    fit_requirements = fit_requirements or FitRequirements()

    if windows.ndim == 0:
        windows = _fit_windows(data, peak_estimates, windows, fit_parameters)

    results = []
    for i in range(windows.sizes[peak_estimates.dim]):
        window = windows[peak_estimates.dim, i]
        data_in_window = data[data.dim, window[0] : window[1]]
        results.append(
            _fit_peak(
                data_in_window,
                window,
                background,
                peak,
                fit_parameters,
                fit_requirements,
            )
        )
    return results


def _fit_peak(
    data: sc.DataArray,
    window: sc.Variable,
    backgrounds: tuple[Model, ...],
    peaks: tuple[Model, ...],
    fit_parameters: FitParameters,
    fit_requirements: FitRequirements,
) -> FitResult:
    candidate_result = None
    # Loop order chosen as [(p0, b0), (p0, b1), (p1, b0), (p1, b1), ...]
    # because trying different background models usually improves fits more than
    # different peak models.
    for peak, background in itertools.product(peaks, backgrounds):
        result = _fit_peak_single_model(
            data,
            peak=peak,
            background=background,
            window=window,
            fit_parameters=fit_parameters,
            fit_requirements=fit_requirements,
        )
        if result.assessment == FitAssessment.success:
            return result
        if candidate_result is None:
            # Keep a potentially failed result to have something to
            # return in case all fits fail.
            candidate_result = result
        # else: reject

    return candidate_result


def _fit_peak_single_model(
    data: sc.DataArray,
    peak: Model,
    background: Model,
    window: sc.Variable,
    fit_parameters: FitParameters,
    fit_requirements: FitRequirements,
) -> FitResult:
    model = background + peak
    bkg_p0 = _guess_background(data, model=background, fit_parameters=fit_parameters)
    p0 = {
        **bkg_p0,
        **_guess_peak(data, model=peak, fit_parameters=fit_parameters),
    }
    bounds = background.param_bounds | _peak_param_bounds(peak)

    if len(data) < len(p0):
        # not enough points to fit all parameters
        return FitResult.for_too_narrow_window(
            peak=peak, background=background, window=window
        )

    bkg_goodness_stats = _fit_background(background, data, bkg_p0)
    try:
        popt, goodness_stats = _perform_fit(model, data, p0=p0, bounds=bounds)
    except RuntimeError as err:
        return FitResult.for_failure(
            peak=peak,
            background=background,
            window=window,
            message=str(err.args[0]),
        )

    assessment = _assess_fit(
        data,
        peak,
        popt,
        goodness_stats,
        bkg_goodness_stats,
        fit_requirements=fit_requirements,
    )
    return FitResult(
        popt=popt,
        assessment=assessment,
        peak=peak,
        background=background,
        window=window,
        message=_message_from_assessment(assessment),
        **goodness_stats,
    )


def _fit_background(
    model: Model, data: sc.DataArray, p0: dict[str, sc.Variable]
) -> dict[str, sc.Variable] | None:
    try:
        _, goodness_stats = _perform_fit(model, data, p0, bounds=model.param_bounds)
    except RuntimeError:
        # Background fits may fail when the background is a bad model.
        # Continue with a background+peak fit instead of aborting.
        return None
    return goodness_stats


def _perform_fit(
    model: Model,
    data: sc.DataArray,
    p0: dict[str, sc.Variable],
    bounds: dict[str, tuple[float, float]],
) -> tuple[dict[str, sc.Variable], dict[str, sc.Variable]]:
    # Workaround for https://github.com/scipp/scipp/issues/3418
    def fit_model(x: sc.Variable, **params: sc.Variable) -> sc.Variable:
        return model(x, **params)

    popt, _ = curve_fit(fit_model, data, p0=p0, bounds=bounds)
    best_fit = sc.DataArray(
        model(data.coords[data.dim], **{k: sc.values(p) for k, p in popt.items()}),
        coords=data.coords,
    )
    goodness_stats = _goodness_of_fit_statistics(data, best_fit, popt)
    return popt, goodness_stats


def _goodness_of_fit_statistics(
    data: sc.DataArray, best_fit: sc.DataArray, params: dict[str, sc.Variable]
) -> dict[str, sc.Variable]:
    # number of degrees of freedom
    n_dof = len(data) - len(params)

    chi_square = _chi_square(data, best_fit)
    reduced_chi_square = chi_square / n_dof
    p = sc.scalar(1 - _scipy_chi2(n_dof).cdf(chi_square.value))

    aic = _akaike_information_criterion(data, chi_square, params)

    return {
        'red_chisq': reduced_chi_square,
        'p_value': p,
        'aic': aic,
    }


def _chi_square(data: sc.DataArray, best_fit: sc.DataArray) -> sc.Variable:
    aux = (sc.values(data) - best_fit) ** 2
    aux /= sc.variances(data)
    return sc.sum(aux.data).to(unit='one')


def _akaike_information_criterion(
    data: sc.DataArray, chi_square: sc.Variable, params: dict
) -> sc.Variable:
    neg2_log_likelihood = len(data) * sc.log(chi_square / len(data))
    return neg2_log_likelihood + 2 * len(params)


def _assess_fit(
    data: sc.DataArray,
    peak: Model,
    popt: dict[str, sc.Variable],
    goodness_stats: dict[str, sc.Variable],
    bkg_goodness_stats: dict[str, sc.Variable] | None,
    fit_requirements: FitRequirements,
) -> FitAssessment:
    """Default fit result assessment.

    Parameters
    ----------
    data:
        Input independent and dependent variable.
    peak:
        Model for the peak.
    popt:
        Optimized parameters.
    goodness_stats:
        Goodness-of-fit statistics for peak + background.
        Includes ``red_chisq``, ``p_value``, ``aic``.
    bkg_goodness_stats:
        Goodness-of-fit statistics for a separate background fit.
        Includes ``red_chisq``, ``p_value``, ``aic``.
    fit_requirements:
        Parameters controlling the fit result assessment.

    Returns
    -------
    :
        Fit assessment.
    """
    if bkg_goodness_stats is not None:
        if bkg_goodness_stats['aic'] < goodness_stats['aic']:
            return FitAssessment.background_is_better
    if (goodness_stats['p_value'] < fit_requirements.min_p_value).value:
        return FitAssessment.p_too_small
    if _peak_is_near_edge(data, popt):
        return FitAssessment.peak_near_edge
    if _curve_points_down(popt):
        return FitAssessment.peak_points_down
    if _peak_is_too_wide(data, peak, popt, fit_requirements):
        return FitAssessment.peak_too_wide
    if _peak_is_too_narrow(data, peak, popt, fit_requirements):
        return FitAssessment.peak_too_narrow
    return FitAssessment.success


def _peak_is_near_edge(data: sc.DataArray, popt: dict[str, sc.Variable]) -> bool:
    coord = data.coords[data.dim]
    step_size = sc.min(coord[1:] - coord[:-1])
    return (popt['peak_loc'] - coord[0] < 2 * step_size).value or (
        coord[-1] - popt['peak_loc'] < 2 * step_size
    ).value


def _curve_points_down(popt: dict[str, sc.Variable]) -> bool:
    try:
        return popt['peak_amplitude'].value < 0
    except KeyError:
        return False


def _peak_is_too_wide(
    data: sc.DataArray,
    peak: Model,
    popt: dict[str, sc.Variable],
    fit_requirements: FitRequirements,
) -> bool:
    fwhm = peak.fwhm(popt)
    coord = data.coords[data.dim]
    return (
        fwhm > fit_requirements.max_peak_width_factor * (coord[-1] - coord[0])
    ).value


def _peak_is_too_narrow(
    data: sc.DataArray,
    peak: Model,
    popt: dict[str, sc.Variable],
    fit_requirements: FitRequirements,
) -> bool:
    fwhm = peak.fwhm(popt)
    coord = data.coords[data.dim]
    center_idx = np.argmin(abs(coord.values - popt['peak_loc'].values))
    # Average of bins around center index.
    # Bins don't normally vary quickly, so this is a good approximation.
    bin_width = (coord[center_idx + 1] - coord[center_idx - 1]) / 2
    return (fwhm < fit_requirements.min_peak_width_factor * bin_width).value


def _guess_background(
    data: sc.DataArray, model: Model, fit_parameters: FitParameters
) -> dict[str, sc.Variable]:
    # 2* because the range is split between beginning and end of window
    n = int(len(data) * fit_parameters.guess_background_fraction / 2)
    tails = sc.concat([data[:n], data[-n:]], dim=data.dim)
    return model.guess(tails)


def _guess_peak(
    data: sc.DataArray, model: Model, fit_parameters: FitParameters
) -> dict[str, sc.Variable]:
    # 2* to match the range in _guess_background
    n = int(len(data) * fit_parameters.guess_background_fraction / 2)
    bulk = data[n:-n]
    return model.guess(bulk)


def _peak_param_bounds(peak: Model) -> dict[str, tuple[float, float]]:
    return {
        **peak.param_bounds,
        'peak_amplitude': (0.0, np.inf),
    }


def _fit_windows(
    data: sc.DataArray,
    center: sc.Variable,
    width: sc.Variable,
    fit_parameters: FitParameters,
) -> sc.Variable:
    windows = sc.empty(sizes={data.dim: len(center), 'range': 2}, unit=center.unit)
    windows['range', 0] = center - width / 2
    windows['range', 1] = np.nextafter(center.values + width.value / 2, np.inf)

    windows = _clip_to_data_range(data, windows)
    _separate_from_neighbors_in_place(center, windows, fit_parameters)

    return windows


def _clip_to_data_range(data: sc.DataArray, windows: sc.Variable) -> sc.Variable:
    lo = data.coords[data.dim].min()
    hi = data.coords[data.dim].max()
    windows = sc.where(windows < lo, lo, windows)
    windows = sc.where(windows > hi, hi, windows)
    return windows


def _separate_from_neighbors_in_place(
    center: sc.Variable, windows: sc.Variable, fit_parameters: FitParameters
) -> None:
    if not sc.issorted(center, center.dim):
        # Needed to easily identify neighbors.
        raise ValueError('Fit window centers must be sorted')

    left_neighbor = center[:-1]
    right_neighbor = center[1:]
    min_separation = (
        right_neighbor - left_neighbor
    ) * fit_parameters.neighbor_separation_factor
    lo = left_neighbor + min_separation
    hi = right_neighbor - min_separation
    # Do not adjust the left edge of the first window and the right edge of the
    # last window because there are no neighbors on those sides.
    left_edge = windows['range', 0][1:]
    right_edge = windows['range', 1][:-1]
    left_edge[:] = sc.where(left_edge < lo, lo, left_edge)
    right_edge[:] = sc.where(right_edge > hi, hi, right_edge)


def _message_from_assessment(assessment: FitAssessment | None) -> str:
    match assessment:
        case FitAssessment.success:
            return 'success'
        case FitAssessment.peak_too_narrow:
            return 'peak too narrow'
        case FitAssessment.peak_too_wide:
            return 'peak too wide'
        case FitAssessment.peak_points_down:
            return 'wrong sign'
        case FitAssessment.window_too_narrow:
            return 'window too narrow'
        case FitAssessment.peak_near_edge:
            return 'too close to edge'
        case FitAssessment.p_too_small:
            return 'p-value too small'
        case FitAssessment.background_is_better:
            return 'background is better'
        case FitAssessment.failed:
            return 'failure'


def _parse_model_spec(
    spec: Model | str | Iterable[Model] | Iterable[str], prefix: str
) -> tuple[Model, ...]:
    if isinstance(spec, Model | str):
        spec = (spec,)
    if not spec:
        raise ValueError(f"No models specified for '{prefix}'")
    return tuple(_parse_single_model_spec(s, prefix=prefix) for s in spec)


def _parse_single_model_spec(spec: Model | str, prefix: str) -> Model:
    match spec:
        case Model():
            return spec.with_prefix(prefix)
        case 'linear':
            return PolynomialModel(degree=1, prefix=prefix)
        case 'quadratic':
            return PolynomialModel(degree=2, prefix=prefix)
        case 'gaussian':
            return GaussianModel(prefix=prefix)
        case 'lorentzian':
            return LorentzianModel(prefix=prefix)
        case 'pseudo_voigt':
            return PseudoVoigtModel(prefix=prefix)
        case _:
            raise ValueError(f"Unknown model: '{spec}'")


def _assert_data_is_supported(data: sc.DataArray) -> None:
    if data.ndim != 1:
        raise sc.DimensionError(
            f'`fit_peaks` requires 1d data, got {data.ndim} dimensions'
        )
    if data.bins is not None:
        raise sc.DTypeError('`fit_peaks` does not support binned data')
    if data.coords.is_edges(data.dim):
        raise sc.CoordError(
            f'Coordinate {data.dim} is bin-edges but `fit_peaks` requires a regular '
            f'coordinate. Consider using `sc.midpoints` to convert the coordinate.'
        )
    if not sc.issorted(data.coords[data.dim], data.dim, order='ascending'):
        # A lot of code here assumes a sorted coord, either to use O(1) instead of O(n)
        # operations or to allow extracting windows.
        raise sc.CoordError(
            'fit_peaks requires the coordinate to be sorted in ascending order. '
            'Consider using scipp.sort to fix this.'
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from collections.abc import Mapping
from typing import Any

import scipp as sc

from .disk_chopper import DiskChopperType


def extract_chopper_from_nexus(
    chopper: Mapping[str, sc.Variable | sc.DataArray | sc.DataGroup],
) -> sc.DataGroup:
    """Convert loaded NeXus disk chopper data to the layout used by ScipNeutron.

    This function extracts relevant time series from ``NXlog``.
    The output may, however, still contain time-dependent fields which need to be
    processed further.
    See :ref:`disk_chopper-time_dependent_parameters`.

    Parameters
    ----------
    chopper:
        The loaded NeXus disk chopper data.

    Returns
    -------
    :
        A new data group with processed fields in the layout expected by
        :meth:`DiskChopper.from_nexus
        <scippneutron.chopper.disk_chopper.DiskChopper.from_nexus>`.
    """
    return sc.DataGroup(
        {
            "type": DiskChopperType(chopper.get("type", DiskChopperType.single)),
            **{key: _parse_field(key, val) for key, val in chopper.items()},
        }
    )


def _parse_field(key: str, value: Any) -> Any:
    if key == "top_dead_center":
        return _parse_tdc(value)
    return _parse_maybe_log(value)


def _parse_tdc(
    tdc: sc.Variable | sc.DataArray | sc.DataGroup,
) -> sc.Variable | sc.DataArray:
    if isinstance(tdc, sc.DataGroup):
        # An NXlog without 'value'
        return tdc["time"]
    return tdc


def _parse_maybe_log(
    x: sc.Variable | sc.DataArray | sc.DataGroup,
) -> sc.Variable | sc.DataArray:
    if isinstance(x, sc.DataGroup) and "value" in x:
        # An NXlog
        return x["value"].squeeze()
    return x



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

from functools import partial

import numpy as np
import pytest
import scipp as sc
import tof as tof_pkg

from scippneutron.chopper import DiskChopper
from scippneutron.conversion.graph.beamline import beamline
from scippneutron.conversion.graph.tof import elastic
from scippneutron.tof import chopper_cascade, fakes, unwrap

sl = pytest.importorskip('sciline')


@pytest.fixture
def disk_choppers():
    wfm1 = DiskChopper(
        frequency=sc.scalar(-70.0, unit="Hz"),
        beam_position=sc.scalar(0.0, unit="deg"),
        phase=sc.scalar(-47.10, unit="deg"),
        axle_position=sc.vector(value=[0, 0, 6.6], unit="m"),
        slit_begin=sc.array(
            dims=["cutout"],
            values=np.array([83.71, 140.49, 193.26, 242.32, 287.91, 330.3]) + 15.0,
            unit="deg",
        ),
        slit_end=sc.array(
            dims=["cutout"],
            values=np.array([94.7, 155.79, 212.56, 265.33, 314.37, 360.0]) + 15.0,
            unit="deg",
        ),
        slit_height=sc.scalar(10.0, unit="cm"),
        radius=sc.scalar(30.0, unit="cm"),
    )

    wfm2 = DiskChopper(
        frequency=sc.scalar(-70.0, unit="Hz"),
        beam_position=sc.scalar(0.0, unit="deg"),
        phase=sc.scalar(-76.76, unit="deg"),
        axle_position=sc.vector(value=[0, 0, 7.1], unit="m"),
        slit_begin=sc.array(
            dims=["cutout"],
            values=np.array([65.04, 126.1, 182.88, 235.67, 284.73, 330.32]) + 15.0,
            unit="deg",
        ),
        slit_end=sc.array(
            dims=["cutout"],
            values=np.array([76.03, 141.4, 202.18, 254.97, 307.74, 360.0]) + 15.0,
            unit="deg",
        ),
        slit_height=sc.scalar(10.0, unit="cm"),
        radius=sc.scalar(30.0, unit="cm"),
    )

    foc1 = DiskChopper(
        frequency=sc.scalar(-56.0, unit="Hz"),
        beam_position=sc.scalar(0.0, unit="deg"),
        phase=sc.scalar(-62.40, unit="deg"),
        axle_position=sc.vector(value=[0, 0, 8.8], unit="m"),
        slit_begin=sc.array(
            dims=["cutout"],
            values=np.array([74.6, 139.6, 194.3, 245.3, 294.8, 347.2]),
            unit="deg",
        ),
        slit_end=sc.array(
            dims=["cutout"],
            values=np.array([95.2, 162.8, 216.1, 263.1, 310.5, 371.6]),
            unit="deg",
        ),
        slit_height=sc.scalar(10.0, unit="cm"),
        radius=sc.scalar(30.0, unit="cm"),
    )

    foc2 = DiskChopper(
        frequency=sc.scalar(-28.0, unit="Hz"),
        beam_position=sc.scalar(0.0, unit="deg"),
        phase=sc.scalar(-12.27, unit="deg"),
        axle_position=sc.vector(value=[0, 0, 15.9], unit="m"),
        slit_begin=sc.array(
            dims=["cutout"],
            values=np.array([98.0, 154.0, 206.8, 255.0, 299.0, 344.65]),
            unit="deg",
        ),
        slit_end=sc.array(
            dims=["cutout"],
            values=np.array([134.6, 190.06, 237.01, 280.88, 323.56, 373.76]),
            unit="deg",
        ),
        slit_height=sc.scalar(10.0, unit="cm"),
        radius=sc.scalar(30.0, unit="cm"),
    )

    return {"wfm1": wfm1, "wfm2": wfm2, "foc1": foc1, "foc2": foc2}


@pytest.mark.parametrize("npulses", [1, 2])
@pytest.mark.parametrize(
    "ltotal",
    [
        sc.array(dims=['detector_number'], values=[26.0], unit='m'),
        sc.array(dims=['detector_number'], values=[26.0, 25.5], unit='m'),
        sc.array(
            dims=['y', 'x'], values=[[26.0, 25.1, 26.33], [25.9, 26.0, 25.7]], unit='m'
        ),
    ],
)
def test_v20_compute_wavelengths_from_wfm(disk_choppers, npulses, ltotal):
    choppers = {
        key: chopper_cascade.Chopper.from_disk_chopper(
            chop, pulse_frequency=sc.scalar(14.0, unit="Hz"), npulses=npulses
        )
        for key, chop in disk_choppers.items()
    }

    monitors = {
        f"detector{i}": ltot for i, ltot in enumerate(ltotal.flatten(to='detector'))
    }

    # Create some neutron events
    wavelengths = sc.array(
        dims=['event'], values=[2.75, 4.2, 5.4, 6.5, 7.6, 8.75], unit='angstrom'
    )
    birth_times = sc.full(sizes=wavelengths.sizes, value=1.5, unit='ms')
    ess_beamline = fakes.FakeBeamlineEss(
        choppers=choppers,
        monitors=monitors,
        run_length=sc.scalar(1 / 14, unit="s") * npulses,
        events_per_pulse=len(wavelengths),
        source=partial(
            tof_pkg.Source.from_neutrons,
            birth_times=birth_times,
            wavelengths=wavelengths,
            frequency=sc.scalar(14.0, unit="Hz"),
        ),
    )

    # Save the true wavelengths for later
    true_wavelengths = ess_beamline.source.data.coords["wavelength"]

    raw_data = sc.concat(
        [ess_beamline.get_monitor(key) for key in monitors.keys()],
        dim='detector',
    ).fold(dim='detector', sizes=ltotal.sizes)

    # Verify that all 6 neutrons made it through the chopper cascade
    assert sc.identical(
        raw_data.bins.concat('pulse').hist().data,
        sc.array(
            dims=['detector'],
            values=[len(wavelengths) * npulses] * len(monitors),
            unit="counts",
            dtype='float64',
        ).fold(dim='detector', sizes=ltotal.sizes),
    )

    # Set up the workflow
    workflow = sl.Pipeline(
        unwrap.unwrap_providers()
        + unwrap.time_of_flight_providers()
        + unwrap.time_of_flight_origin_from_choppers_providers(wfm=True)
    )
    workflow[unwrap.PulsePeriod] = sc.reciprocal(ess_beamline.source.frequency)
    workflow[unwrap.PulseStride | None] = None

    # Define the extent of the pulse that contains the 6 neutrons in time and wavelength
    # Note that we make a larger encompassing pulse to ensure that the frame bounds are
    # computed correctly
    workflow[unwrap.SourceTimeRange] = (
        sc.scalar(0.0, unit='ms'),
        sc.scalar(3.4, unit='ms'),
    )
    workflow[unwrap.SourceWavelengthRange] = (
        sc.scalar(0.2, unit='angstrom'),
        sc.scalar(10.0, unit='angstrom'),
    )

    workflow[unwrap.Choppers] = choppers
    workflow[unwrap.Ltotal] = ltotal
    workflow[unwrap.RawData] = raw_data

    # Compute time-of-flight
    tofs = workflow.compute(unwrap.TofData)
    assert {dim: tofs.sizes[dim] for dim in ltotal.sizes} == ltotal.sizes

    # Convert to wavelength
    graph = {**beamline(scatter=False), **elastic("tof")}
    wav_wfm = tofs.transform_coords("wavelength", graph=graph)

    # Compare the computed wavelengths to the true wavelengths
    for i in range(npulses):
        result = wav_wfm['pulse', i].flatten(to='detector')
        for j in range(len(result)):
            computed_wavelengths = result[j].values.coords["wavelength"]
            assert sc.allclose(
                computed_wavelengths,
                true_wavelengths['pulse', i],
                rtol=sc.scalar(1e-02),
            )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

"""Peak fitting and removal.

.. rubric:: Classes

.. autosummary::
  :toctree:
  :template: class-template.rst

  FitAssessment
  FitParameters
  FitResult
  FitRequirements

.. rubric:: Functions

.. autosummary::
  :toctree:

  fit_peaks
  remove_peaks
"""

from ._common import FitParameters, FitRequirements
from ._fit_peaks import FitAssessment, FitResult, fit_peaks
from ._remove_peaks import remove_peaks

__all__ = [
    'FitAssessment',
    'FitParameters',
    'FitRequirements',
    'FitResult',
    'fit_peaks',
    'remove_peaks',
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Jan-Lukas Wynen

import pytest
import scipp as sc

import scippneutron as scn

try:
    import mantid.kernel as kernel
    import mantid.simpleapi as sapi
except ImportError:
    pytestmark = pytest.mark.skip('Mantid framework is unavailable')
    sapi = None
    kernel = None


def mantid_coord(scipp_coord: str) -> str:
    return {
        'tof': 'TOF',
        'wavelength': 'Wavelength',
        'dspacing': 'dSpacing',
        'energy': 'Energy',
        'energy_transfer': 'DeltaE',
    }[scipp_coord]


def make_workspace(
    coord: str, emode: str = 'Elastic', efixed: sc.Variable | None = None
) -> 'sapi.Workspace':
    ws = sapi.CreateSampleWorkspace(XMin=1000, NumBanks=1, StoreInADS=False)
    # Crop out spectra index 0 as has two_theta=0, gives inf d-spacing
    ws = sapi.CropWorkspace(ws, StartWorkspaceIndex=1, StoreInADS=False)
    ws = sapi.ConvertUnits(
        InputWorkspace=ws,
        Target=mantid_coord(coord),
        EMode=emode,
        EFixed=efixed.value if efixed is not None else None,
        StoreInADS=False,
    )  # start in origin units
    ws.mutableRun().addProperty(
        'deltaE-mode', kernel.StringPropertyWithValue('deltaE-mode', emode), '', True
    )
    if efixed is not None:
        ws.mutableRun().addProperty(
            'Ei',
            kernel.FloatPropertyWithValue('Ei', efixed.value),
            str(efixed.unit),
            False,
        )
    return ws


def mantid_convert_units(
    ws: 'sapi.Workspace',
    target: str,
    emode: str = 'Elastic',
    efixed: sc.Variable | None = None,
) -> sc.DataArray:
    out_ws = sapi.ConvertUnits(
        InputWorkspace=ws,
        Target=mantid_coord(target),
        EMode=emode,
        EFixed=efixed.value if efixed is not None else None,
        StoreInADS=False,
    )
    out = scn.mantid.from_mantid(out_ws)['data']
    # broadcast to circumvent common-bins conversion in from_mantid
    spec_shape = out.coords['spectrum'].shape
    out.coords[target] = (
        sc.ones(dims=['spectrum'], shape=spec_shape) * out.coords[target]
    )
    return out


def test_mantid_convert_tof_to_wavelength():
    in_ws = make_workspace('tof')
    out_mantid = mantid_convert_units(in_ws, 'wavelength')

    in_da = scn.mantid.from_mantid(in_ws)['data']
    out_scipp = scn.convert(data=in_da, origin='tof', target='wavelength', scatter=True)

    assert sc.allclose(
        out_scipp.coords['wavelength'],
        out_mantid.coords['wavelength'],
        rtol=1e-8 * sc.units.one,
    )
    assert sc.identical(out_scipp.coords['spectrum'], out_mantid.coords['spectrum'])


def test_mantid_convert_tof_to_dspacing():
    in_ws = make_workspace('tof')
    out_mantid = mantid_convert_units(in_ws, 'dspacing')

    in_da = scn.mantid.from_mantid(in_ws)['data']
    out_scipp = scn.convert(data=in_da, origin='tof', target='dspacing', scatter=True)

    assert sc.allclose(
        out_scipp.coords['dspacing'],
        out_mantid.coords['dspacing'],
        rtol=1e-8 * sc.units.one,
    )
    assert sc.identical(out_scipp.coords['spectrum'], out_mantid.coords['spectrum'])


def test_mantid_convert_tof_to_energy():
    in_ws = make_workspace('tof')
    out_mantid = mantid_convert_units(in_ws, 'energy')

    in_da = scn.mantid.from_mantid(in_ws)['data']
    out_scipp = scn.convert(data=in_da, origin='tof', target='energy', scatter=True)

    # Mantid reverses the order of the energy dim.
    mantid_energy = sc.empty_like(out_mantid.coords['energy'])
    assert mantid_energy.dims[1] == 'energy'
    mantid_energy.values = out_mantid.coords['energy'].values[..., ::-1]

    assert sc.allclose(
        out_scipp.coords['energy'], mantid_energy, rtol=1e-7 * sc.units.one
    )
    assert sc.identical(out_scipp.coords['spectrum'], out_mantid.coords['spectrum'])


def test_mantid_convert_tof_to_direct_energy_transfer():
    efixed = 1000.0 * sc.Unit('meV')
    in_ws = make_workspace('tof', emode='Direct', efixed=efixed)
    out_mantid = mantid_convert_units(
        in_ws, 'energy_transfer', emode='Direct', efixed=efixed
    )

    in_da = scn.mantid.from_mantid(in_ws)['data']
    out_scipp = scn.convert(
        data=in_da, origin='tof', target='energy_transfer', scatter=True
    )

    # The conversion consists of multiplications and additions, thus the relative error
    # changes with the inputs. In this case, small tof yields a large error due to
    # the 1/tof**2 factor in the conversion.
    # rtol is chosen to account for linearly changing tof in the input data.
    assert sc.allclose(
        out_scipp.coords['energy_transfer'],
        out_mantid.coords['energy_transfer'],
        rtol=sc.linspace(
            'energy_transfer',
            1e-6,
            1e-10,
            out_scipp.coords['energy_transfer'].sizes['energy_transfer'],
        ),
    )
    assert sc.identical(out_scipp.coords['spectrum'], out_mantid.coords['spectrum'])



from functools import partial
from typing import Any

import scipp as sc

from .material import Material
from .types import SampleShape


def compute_transmission_map(
    sample_shape: SampleShape,
    sample_material: Material,
    beam_direction: sc.Variable,
    wavelength: sc.Variable,
    detector_position: sc.Variable,
    quadrature_kind: Any = 'medium',
) -> sc.DataArray:
    points, weights = sample_shape.quadrature(quadrature_kind)
    transmission = _integrate_transmission_fraction(
        partial(
            _single_scatter_distance_through_sample,
            sample_shape,
            points,
            beam_direction,
        ),
        partial(_transmission_fraction, sample_material),
        points,
        weights,
        detector_position,
        wavelength,
    )
    return sc.DataArray(
        data=transmission / sample_shape.volume,
        coords={'detector_position': detector_position, 'wavelength': wavelength},
    )


def _single_scatter_distance_through_sample(
    sample_shape, scatter_point, initial_direction, scatter_direction
):
    L1 = sample_shape.beam_intersection(scatter_point, -initial_direction)
    L2 = sample_shape.beam_intersection(scatter_point, scatter_direction)
    return L1 + L2


def _transmission_fraction(material, distance_through_sample, wavelength):
    return sc.exp(
        -(material.attenuation_coefficient(wavelength) * distance_through_sample).to(
            unit='dimensionless', copy=False
        )
    )


def _integrate_transmission_fraction(
    distance_through_sample,
    transmission,
    points,
    weights,
    detector_position,
    wavelengths,
):
    # If size after broadcast is too large
    # then don't vectorize the operation to avoid OOM
    if points.size * detector_position.size > 20_000_000:
        out = []
        dim = detector_position.dims[0]
        for i in range(detector_position.sizes[dim]):
            out.append(  # noqa: PERF401
                _integrate_transmission_fraction(
                    distance_through_sample,
                    transmission,
                    points,
                    weights,
                    detector_position[dim, i],
                    wavelengths,
                )
            )

        return sc.concat(
            out,
            dim=dim,
        )

    scatter_direction = detector_position - points.to(unit=detector_position.unit)
    scatter_direction /= sc.norm(scatter_direction)
    Ltot = distance_through_sample(scatter_direction)

    # The Ltot array is already large, to avoid OOM, don't vectorize this operation
    return sc.concat(
        [
            # Instead of broadcast multiply and sum, use matvec for efficiency
            # this becomes relevant when the number of wavelength points grows
            sc.array(
                dims=(tf := transmission(Ltot, w)).dims[:-1],
                values=tf.values @ weights.values,
                unit=tf.unit * weights.unit,
            )
            for w in wavelengths
        ],
        dim=wavelengths.dim,
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Jan-Lukas Wynen
r"""Functions for computing coordinates related to beamline geometry.

Most functions in this module assume a straight beamline geometry.
That is, the beams are not curved by, e.g., a beam guide.
Specialized functions are provided for handling gravity.

ScippNeutron uses three positions to define a beamline:

- ``source_position`` defines the point where :math:`t=0` (or vice versa).
  In practice, this can be the actual neutron source, a moderator, or a chopper.
- ``sample_position`` is the position of the sample that scatters neutrons.
- ``position`` is the position of the detector (pixel / voxel) that detects a neutron
  at :math:`t=\mathsf{tof}`.

Based on these positions, we define:

- ``incident_beam`` is the vector of incoming neutrons on the sample.
  It points from the source to the sample.
  (:func:`straight_incident_beam`)
- ``L1`` (or :math:`L_1`) is the length of ``incident_beam``.
  (:func:`L1`)
- ``scattered_beam`` is the vector of neutrons that were scattered off the sample.
  It points from the sample to the detector pixels.
  (:func:`straight_scattered_beam`)
- ``L2`` (or :math:`L_2`) is the length of ``scattered_beam``.
  (:func:`L2`)
- ``Ltotal`` is the total beam length :math:`L_\mathsf{total} = L_1 + L_2`.
  (:func:`total_beam_length` and :func:`total_straight_beam_length_no_scatter`)

Coordinate system
-----------------

Note
----
  The coordinate system is not needed by most operations because beamline coordinates
  are usually relative to the positions.
  As long as ``position``, ``source_position``, and ``sample_position`` are defined
  consistently, ``incident_beam``, ``scattered_beam``, and ``two_theta`` can be
  computed without knowledge of the coordinate system.

  However, we need the actual coordinate system to compute ``phi``.

ScippNeutron uses a coordinate system aligned with
the incident beam and gravity.
ScippNeutron's coordinate system corresponds to that of
`NeXus <https://manual.nexusformat.org/design.html#the-nexus-coordinate-system>`_
when the incident beam is perpendicular to gravity.
The image below shows how coordinates are defined with respect to the
quantities defined above.
The plot on the right-hand side shows the view from the sample along the :math:`z`-axis,
that is, the :math:`z`-axis points towards the viewer.
(The sample is placed in the origin here; this is only done for illustration purposes
and not required.)

.. image:: ../../../docs/_static/beamline/beamline_coordinates_light.svg
   :class: only-light
   :scale: 100 %
   :alt: Beamline coordinate system.
   :align: center

.. image:: ../../../docs/_static/beamline/beamline_coordinates_dark.svg
   :class: only-dark
   :scale: 100 %
   :alt: Beamline coordinate system.
   :align: center

The axes are defined by these unit vectors:

.. math::

    \hat{e}_y &= -g / |g| \\
    z_{\text{proj}} &= b_1 - (b_1 \cdot \hat{e}_y) \hat{e}_y \\
    \hat{e}_z &= z_{\text{proj}} / |z_{\text{proj}}| \\
    \hat{e}_x &= \hat{e}_y \times \hat{e}_z

which span an orthogonal, right-handed coordinate system.
Here, :math:`b_1` is the ``incident_beam`` and :math:`g` is the gravity vector.
This means that the y-axis is antiparallel to gravity.
The z-axis is defined by projecting the incident beam onto a plane perpendicular
to gravity.
Basis vectors can be computed using :func:`beam_aligned_unit_vectors`.

:math:`p = \sqrt{x^2 + y^2}` is the projection of the
scattered beam onto the :math:`x-y` plane.
It is included here because it is used by some coordinate transformations.

The scattering angles are defined similarly to spherical coordinates as:

.. math ::

    \mathsf{cos}(2\theta) &= \frac{b_1 \cdot b_2}{|b_1| |b_2|} \\
    \mathsf{tan}(\phi) &= \frac{b_2 \cdot \hat{e}_y}{b_2 \cdot \hat{e}_x}

where :math:`b_2` is the scattered beam.

- ``two_theta`` is the angle between the scattered beam and incident beam.
  Note the extra factor 2 compared to spherical coordinates;
  it ensures that the definition corresponds to Bragg's law.
- ``phi`` is the angle between the x-axis and the projection of the scattered beam
  onto the x-y-plane.

These definitions assume that gravity can be neglected.
See :func:`scattering_angles_with_gravity` for definitions that account for gravity.
And :func:`scattering_angle_in_yz_plane` for the definition used in reflectometry ---
which also includes gravity.
"""

from typing import TypedDict

import scipp as sc
import scipp.constants
from scipp.typing import VariableLike

from .._utils import elem_dtype, elem_unit


def L1(*, incident_beam: VariableLike) -> VariableLike:
    """Compute the length of the incident beam.

    The result is the primary beam length

    .. math::

        L_1 = | \\mathtt{incident\\_beam} |

    Parameters
    ----------
    incident_beam:
        Beam from source to sample. Expects ``dtype=vector3``.

    Returns
    -------
    :
        :math:`L_1`

    See Also
    --------
    straight_incident_beam:
        Compute the incident beam for a straight beamline.
    """
    return sc.norm(incident_beam)


def L2(*, scattered_beam: VariableLike) -> VariableLike:
    """Compute the length of the scattered beam.

    The result is the secondary beam length

    .. math::

        L_2 = | \\mathtt{incident\\_beam} |

    Parameters
    ----------
    scattered_beam:
        Beam from sample to detector. Expects ``dtype=vector3``.

    Returns
    -------
    :
        :math:`L_2`

    See Also
    --------
    straight_scattered_beam:
        Compute the scattered beam for a straight beamline.
    """
    return sc.norm(scattered_beam)


def straight_incident_beam(
    *, source_position: VariableLike, sample_position: VariableLike
) -> VariableLike:
    """Compute the length of the beam from source to sample.

    Assumes a straight beam.
    The result is

    .. math::

        \\mathtt{incident\\_beam} = \\mathtt{sample\\_position}
                                    - \\mathtt{source\\_position}

    Parameters
    ----------
    source_position:
        Position of the beam's source.
    sample_position:
        Position of the sample.

    Returns
    -------
    :
        ``incident_beam``
    """
    return sample_position - source_position


def straight_scattered_beam(
    *, position: VariableLike, sample_position: VariableLike
) -> VariableLike:
    """Compute the length of the beam from sample to detector.

    Assumes a straight beam.
    The result is

    .. math::

        \\mathtt{scattered\\_beam} = \\mathtt{position} - \\mathtt{sample\\_position}

    Parameters
    ----------
    position:
        Position of the detector.
    sample_position:
        Position of the sample.

    Returns
    -------
    :
        ``scattered_beam``
    """
    return position - sample_position


def total_beam_length(*, L1: VariableLike, L2: VariableLike) -> VariableLike:
    """Compute the combined length of the incident and scattered beams.

    The result is

    .. math::

        L_\\mathsf{total} = | L_1 + L_2 |

    Parameters
    ----------
    L1:
        Primary path length (incident beam).
    L2:
        Secondary path length (scattered beam).

    Returns
    -------
    :
        :math:`L_\\mathsf{total}`
    """
    return L1 + L2


def total_straight_beam_length_no_scatter(
    *, source_position: VariableLike, position: VariableLike
) -> VariableLike:
    """Compute the length of the beam from source to given position.

    Assumes a straight beam.
    The result is

    .. math::

        L_\\mathsf{total} = | \\mathtt{position} - \\mathtt{{source\\_position}} |

    Parameters
    ----------
    source_position:
        Position of the beam's source. Expects ``dtype=vector3``.
    position:
        Position of the detector. Expects ``dtype=vector3``.

    Returns
    -------
    :
        :math:`L_\\mathsf{total}`
    """
    return sc.norm(position - source_position)


def two_theta(
    *, incident_beam: VariableLike, scattered_beam: VariableLike
) -> VariableLike:
    """Compute the scattering angle between scattered and transmitted beams.

    See :mod:`beamline` for the definition of the angle.

    The result is equivalent to

    .. math::

        b_1 &= \\mathtt{incident\\_beam} / |\\mathtt{incident\\_beam}| \\\\
        b_2 &= \\mathtt{scattered\\_beam} / |\\mathtt{scattered\\_beam}| \\\\
        2\\theta &= \\mathsf{acos}(b_1 \\cdot b_2)

    but uses a numerically more stable implementation by W. Kahan
    described in paragraph of :cite:`kahan:2000:CPR`.
    See also
    https://people.eecs.berkeley.edu/~wkahan/MathH110/Cross.pdf.

    Parameters
    ----------
    incident_beam:
        Beam from source to sample. Expects ``dtype=vector3``.
    scattered_beam:
        Beam from sample to detector. Expects ``dtype=vector3``.

    Returns
    -------
    :
        :math:`2\\theta`

    See Also
    --------
    straight_incident_beam:
        Compute the incident beam for a straight beamline.
    straight_scattered_beam:
        Compute the scattered beam for a straight beamline.
    scattering_angles_with_gravity:
        Calculate ``two_theta`` and ``phi`` with gravity.
    """
    # The implementation is based on paragraph 13 of
    # https://people.eecs.berkeley.edu/~wkahan/MathH110/Cross.pdf
    # It claims that the formula is 'valid for euclidean spaces of any dimension'
    # and 'it never errs by more than a modest multiple of epsilon'
    # where 'epsilon is the roundoff threshold for individual arithmetic
    # operations'.
    b1 = incident_beam / L1(incident_beam=incident_beam)
    b2 = scattered_beam / L2(scattered_beam=scattered_beam)

    y = sc.norm(b1 - b2)
    b2 += b1
    x = sc.norm(b2)
    res = sc.atan2(y=y, x=x, out=x)
    res *= 2
    return res


class BeamAlignedUnitVectors(TypedDict):
    """A dict with keys 'beam_aligned_unit_{x,y,z}'."""

    beam_aligned_unit_x: sc.Variable
    """Unit vector in x-direction."""
    beam_aligned_unit_y: sc.Variable
    """Unit vector in y-direction."""
    beam_aligned_unit_z: sc.Variable
    """Unit vector in z-direction."""


def beam_aligned_unit_vectors(
    incident_beam: sc.Variable, gravity: sc.Variable
) -> BeamAlignedUnitVectors:
    r"""Return unit vectors for a coordinate system aligned with the incident beam.

    The unit vectors are

    .. math::

        \hat{e}_y &= -g / |g| \\
        z_{\text{proj}} &= b_1 - (b_1 \cdot \hat{e}_y) \hat{e}_y \\
        \hat{e}_z &= z_{\text{proj}} / |z_{\text{proj}}| \\
        \hat{e}_x &= \hat{e}_y \times \hat{e}_z

    where :math:`b_1` is the ``incident_beam`` and :math:`g` is the gravity vector.
    See the module-level docs of :mod:`scippneutron.conversion.beamline` for  details.

    Parameters
    ----------
    incident_beam:
        Beam from source to sample. Expects ``dtype=vector3``.
    gravity:
        Gravity vector. Expects ``dtype=vector3``.

    Returns
    -------
    A dict containing the unit vectors with keys 'ex', 'ey', and 'ez'.

    Raises
    ------
    ValueError
        If the incident beam is parallel to gravity.
        The rotation of the x-z plane is ill-define din this case.

    See Also
    --------
    straight_incident_beam:
        Compute the incident beam for a straight beamline.
    """
    ey = -gravity / sc.norm(gravity)
    # Project incident_beam onto a plane perpendicular to ey.
    z = incident_beam - sc.dot(incident_beam, ey) * ey
    z_norm = sc.norm(z)
    if sc.any(z_norm < sc.scalar(1e-10, unit=z_norm.unit)):
        raise ValueError(
            "Cannot construct a coordinate system. The incident beam and "
            "gravity are parallel to each other."
        )
    ez = z / sc.norm(z)
    ex = sc.cross(ey, ez)
    return {
        'beam_aligned_unit_x': ex,
        'beam_aligned_unit_y': ey,
        'beam_aligned_unit_z': ez,
    }


def _drop_due_to_gravity(
    distance: sc.Variable,
    wavelength: sc.Variable,
    gravity: sc.Variable,
) -> sc.Variable:
    """Compute the distance a neutron drops due to gravity.

    See the documentation of :func:`scattering_angles_with_gravity`.
    """
    distance = distance.to(dtype=elem_dtype(wavelength), copy=False)
    const = (sc.norm(gravity) * (sc.constants.m_n**2 / (2 * sc.constants.h**2))).to(
        dtype=elem_dtype(wavelength), copy=False
    )

    # Convert unit to eventually match the unit of y.
    # Copy to make it safe to use in-place ops.
    drop = wavelength.to(
        unit=sc.sqrt(sc.reciprocal(elem_unit(distance) * elem_unit(const))), copy=True
    )
    drop *= drop
    drop *= const

    distance *= distance
    if set(distance.dims).issubset(drop.dims):
        drop *= distance
        return drop
    return drop * distance


class SphericalCoordinates(TypedDict):
    """A dict with keys 'two_theta' and 'phi'."""

    two_theta: sc.Variable
    """Polar angle.

    Between the scattered beam and continuation of the incident beam.
    Mind the factor 2 which is defined as in Bragg's law.
    """

    phi: sc.Variable
    """Azimuthal angle.

    The angle between the projection of the scattered beam onto the x-y plane
    and the x-axis.
    """


def scattering_angles_with_gravity(
    incident_beam: sc.Variable,
    scattered_beam: sc.Variable,
    wavelength: sc.Variable,
    gravity: sc.Variable,
) -> SphericalCoordinates:
    r"""Compute scattering angles theta and phi using gravity.

    With the definitions of the unit vectors in
    `Coordinate system <./scippneutron.conversion.beamline.rst#coordinate-system>`_,
    we have the components of the scattered beam :math:`b_2` in the beam-aligned
    coordinate system:

    .. math::

        x_d &= b_2 \cdot \hat{e}_x \\
        y_d &= b_2 \cdot \hat{e}_y \\
        z_d &= b_2 \cdot \hat{e}_z

    :math:`b_2` points from the sample to the detector that detected the neutron.
    The neutron left the sample in the direction of a vector :math:`b'_2`.
    This vector defines the scattering angles :math:`2\theta` and :math:`\phi`.
    Taking gravity into account, we have :math:`b'_2 \neq b_2`.
    Solving the equations of motion gives the following for the
    components of :math:`b'_2`:

    .. math::

        x'_d &= x_d \\
        y'_d &= y_d + \delta_y \\
        z'_d &= z_d

    Where :math:`|g|` is the strength of gravity, :math:`m_n` is the neutron mass,
    :math:`h` is the Planck constant, :math:`\lambda` is the wavelength, and

    .. math::

        \delta_y = \frac{|g| m_n^2}{2 h^2} L_2^{\prime\, 2} \lambda^2

    This gives the gravity-corrected scattering angles:

    .. math::

        2\theta &= \sphericalangle(b_1, b_2 + \delta_y \hat{e}_y) \\
        \mathsf{tan}(\phi) &= \frac{y'_d}{x_d}

    where :math:`\sphericalangle` is the angle between two vectors as implemented by
    :func:`two_theta`.
    When :math:`b_1` is orthogonal to gravity, we can equivalently use

    .. math::

        \mathsf{tan}(2\theta) = \frac{\sqrt{x_d^2 + y_d^{\prime\, 2}}}{z_d}

    This equation allowed for a more efficient implementation.

    Attention
    ---------
        The above equation for :math:`y'_d` contains :math:`L_2^{\prime\, 2} = |b'_2|`
        which in turn depends on :math:`y'_d`.
        Solving this equation for :math:`y'_d` is too difficult.
        Instead, we approximate :math:`L'_2 \approx L_2`.
        The impact of this approximation on :math:`2\theta` is of the order of
        :math:`10^{-6}` or less for beamlines at ESS.
        This is within the expected statistical uncertainties and can be ignored.

        See `two_theta gravity correction
        <../../user-guide/algorithms-background/two_theta-gravity-correction.rst>`_
        for details.

    Parameters
    ----------
    incident_beam:
        Beam from source to sample. Expects ``dtype=vector3``.
    scattered_beam:
        Beam from sample to detector. Expects ``dtype=vector3``.
    wavelength:
        Wavelength of neutrons.
    gravity:
        Gravity vector.

    Returns
    -------
    :
        A dict containing the polar scattering angle ``'two_theta'`` and
        the azimuthal angle ``'phi'``.

    See also
    --------
    scattering_angle_in_yz_plane:
        Ignores the ``x`` component when computing ``theta``.
        This is used in reflectometry.
    """
    if sc.any(
        abs(sc.dot(gravity, incident_beam))
        > sc.scalar(1e-10, unit=incident_beam.unit) * sc.norm(gravity)
    ):
        return _scattering_angles_with_gravity_generic(
            incident_beam=incident_beam,
            scattered_beam=scattered_beam,
            wavelength=wavelength,
            gravity=gravity,
        )
    return _scattering_angles_with_gravity_orthogonal_coords(
        incident_beam=incident_beam,
        scattered_beam=scattered_beam,
        wavelength=wavelength,
        gravity=gravity,
    )


def _scattering_angles_with_gravity_generic(
    incident_beam: sc.Variable,
    scattered_beam: sc.Variable,
    wavelength: sc.Variable,
    gravity: sc.Variable,
) -> SphericalCoordinates:
    unit_vectors = beam_aligned_unit_vectors(
        incident_beam=incident_beam, gravity=gravity
    )
    ex = unit_vectors['beam_aligned_unit_x']
    ey = unit_vectors['beam_aligned_unit_y']

    drop_distance = _drop_due_to_gravity(
        distance=sc.norm(scattered_beam), wavelength=wavelength, gravity=gravity
    )
    y = drop_distance + sc.dot(scattered_beam, ey).to(
        dtype=elem_dtype(wavelength), copy=False
    )
    x = sc.dot(scattered_beam, ex).to(dtype=elem_dtype(y), copy=False)
    phi = sc.atan2(y=y, x=x, out=y)

    drop = drop_distance * (gravity / sc.norm(gravity))
    drop += scattered_beam
    return {
        'two_theta': two_theta(incident_beam=incident_beam, scattered_beam=drop).to(
            dtype=elem_dtype(wavelength), copy=False
        ),
        'phi': phi,
    }


# This is an optimized implementation that only works when `incident_beam` and `gravity`
# are orthogonal to each other. It uses less memory than the generic version.
def _scattering_angles_with_gravity_orthogonal_coords(
    incident_beam: sc.Variable,
    scattered_beam: sc.Variable,
    wavelength: sc.Variable,
    gravity: sc.Variable,
) -> SphericalCoordinates:
    unit_vectors = beam_aligned_unit_vectors(
        incident_beam=incident_beam, gravity=gravity
    )
    ex = unit_vectors['beam_aligned_unit_x']
    ey = unit_vectors['beam_aligned_unit_y']
    ez = unit_vectors['beam_aligned_unit_z']

    y = _drop_due_to_gravity(
        distance=sc.norm(scattered_beam), wavelength=wavelength, gravity=gravity
    )
    y += sc.dot(scattered_beam, ey).to(dtype=elem_dtype(wavelength), copy=False)

    x = sc.dot(scattered_beam, ex).to(dtype=elem_dtype(y), copy=False)
    phi = sc.atan2(y=y, x=x)

    # Corresponds to `two_theta_ = sc.atan2(y=sc.sqrt(x**2 + y**2), x=z)`
    x *= x
    y *= y
    y += x
    del x
    y = sc.sqrt(y, out=y)
    z = sc.dot(scattered_beam, ez).to(dtype=elem_dtype(y), copy=False)
    two_theta_ = sc.atan2(y=y, x=z, out=y)

    return {'two_theta': two_theta_, 'phi': phi}


def scattering_angle_in_yz_plane(
    incident_beam: sc.Variable,
    scattered_beam: sc.Variable,
    wavelength: sc.Variable,
    gravity: sc.Variable,
) -> sc.Variable:
    r"""Compute polar scattering angles in the y-z plane using gravity.

    Note
    ----
        This function uses the reflectometry definition of the polar scattering angle.
        Other techniques define the angle w.r.t. the incident beam.
        See :func:`scattering_angles_with_gravity` for those use cases.

    With the definitions given in :func:`scattering_angles_with_gravity`,
    and ignoring :math:`x_d`, we get

    .. math::

        \mathsf{tan}(\gamma) = \frac{|y_d^{\prime}|}{z_d}

    with

    .. math::

        y'_d = y_d + \frac{|g| m_n^2}{2 h^2} L_2^{\prime\, 2} \lambda^2

    The angle :math:`\gamma` is defined as in Fig. 5 of :cite:`STAHN201644`.

    Attention
    ---------
        The above equation for :math:`y'_d` contains :math:`L_2^{\prime\, 2} = |b'_2|`
        which in turn depends on :math:`y'_d`.
        Solving this equation for :math:`y'_d` is too difficult.
        Instead, we approximate :math:`L'_2 \approx L_2`.
        The impact of this approximation on :math:`\gamma` is of the order of
        :math:`10^{-6}` or less for beamlines at ESS.
        This is within the expected statistical uncertainties and can be ignored.

        See `two_theta gravity correction
        <../../user-guide/algorithms-background/two_theta-gravity-correction.rst>`_
        for details.

    Parameters
    ----------
    incident_beam:
        Beam from source to sample. Expects ``dtype=vector3``.
    scattered_beam:
        Beam from sample to detector. Expects ``dtype=vector3``.
    wavelength:
        Wavelength of neutrons.
    gravity:
        Gravity vector.

    Returns
    -------
    :
        The polar scattering angle :math:`\gamma`.

    See also
    --------
    scattering_angles_with_gravity:
        Includes the ``x`` component when computing ``theta``.
        This is used in techniques other than reflectometry.
    """
    if sc.any(
        abs(sc.dot(gravity, incident_beam))
        > sc.scalar(1e-10, unit=incident_beam.unit) * sc.norm(gravity)
    ):
        raise ValueError(
            '`gravity` and `incident_beam` must be orthogonal. '
            f'Got a deviation of {sc.dot(gravity, incident_beam).max():c}. '
            'It is unclear how the angle is defined in this case.'
        )

    unit_vectors = beam_aligned_unit_vectors(
        incident_beam=incident_beam, gravity=gravity
    )
    ey = unit_vectors['beam_aligned_unit_y']
    ez = unit_vectors['beam_aligned_unit_z']

    y = _drop_due_to_gravity(
        distance=sc.norm(scattered_beam), wavelength=wavelength, gravity=gravity
    )
    y += sc.dot(scattered_beam, ey).to(dtype=elem_dtype(wavelength), copy=False)
    y = sc.abs(y, out=y)
    z = sc.dot(scattered_beam, ez).to(dtype=elem_dtype(y), copy=False)
    return sc.atan2(y=y, x=z, out=y)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
from __future__ import annotations

from dataclasses import dataclass


@dataclass(kw_only=True, slots=True)
class FitParameters:
    """Parameters for fitting peaks."""

    guess_background_fraction: float = 0.5
    """The fraction of fit windows used for estimating the background.

    Determines what fraction of each fit window is used for estimating the initial
    parameters for background fits.
    E.g., a value of 0.5 means that half the range is used for the background;
    specifically first and last quarter of the window.
    Initial parameters for the peak are estimated from the remainder of the window.
    """
    neighbor_separation_factor: float = 1 / 3
    r"""Determines how close fit windows may be to neighboring peaks.

    The fit window for a peak with initial estimate :math:`p_i` is shrunk to the
    exclusive interval :math:`(p_i - \Delta, p_i + \Delta)` if it is smaller
    than the given fit window, where

    .. math::

        \Delta = (p_{i+1} - p_{i-1}) \cdot \mathsf{neighbor\_separation\_factor}

    is based on the separation between the left and right neighbor estimate of the peak.

    Importantly, this uses the peak *estimates*, not the actual fitted peak positions.
    """


@dataclass(kw_only=True, slots=True)
class FitRequirements:
    """Requirements that fitted models must satisfy to be considered successful."""

    min_p_value: float = 0.01
    """Minimum for the p-value of the fit.

    See :attr:`FitResult.p_value`.
    """
    max_peak_width_factor: float = 1.0
    r"""Maximum allowed width of peaks relative to fit window.

    The full width at half maximum of fitted peaks must satisfy

    .. math::

        \mathsf{FWHM} < \mathsf{max\_peak\_width\_factor} \cdot \mathsf{window\_width}
    """
    min_peak_width_factor: float = 1.0
    r"""Minimum allowed width of peaks relative to coordinate spacing.

    The full width at half maximum of fitted peaks must satisfy

    .. math::

        \mathsf{FWHM} < \mathsf{min\_peak\_width\_factor} \cdot \Delta_{\mathsf{coord}}

    where :math:`\Delta_{\mathsf{coord}}` is the spacing of the coordinate around
    the peak center.
    """



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

import pytest
import scipp as sc
import scipp.testing

from scippneutron.chopper import collapse_plateaus, filter_in_phase, find_plateaus


def test_find_plateaus_only_plateau_exact_float():
    da = sc.DataArray(
        sc.full(value=4.2, sizes={'x': 12}), coords={'x': sc.arange('x', 12)}
    )
    plateaus = find_plateaus(da, atol=sc.scalar(1e-7), min_n_points=3)
    assert plateaus.sizes == {'plateau': 1}
    sc.testing.assert_identical(
        plateaus.coords['plateau'], sc.array(dims=['plateau'], values=[0], unit=None)
    )
    sc.testing.assert_identical(plateaus[0].value, da)


def test_find_plateaus_only_plateau_exact_int():
    da = sc.DataArray(
        sc.full(value=6, sizes={'x': 12}, dtype='int64'),
        coords={'x': sc.arange('x', 12)},
    )
    plateaus = find_plateaus(da, atol=sc.scalar(1e-7), min_n_points=3)
    assert plateaus.sizes == {'plateau': 1}
    sc.testing.assert_identical(
        plateaus.coords['plateau'], sc.array(dims=['plateau'], values=[0], unit=None)
    )
    sc.testing.assert_identical(plateaus[0].value, da)


def test_find_plateaus_only_plateau_approx():
    da = sc.DataArray(
        sc.array(dims=['y'], values=[4.02, 4.05, 3.97, 4.0], unit='m'),
        coords={'y': sc.arange('y', 4, unit='s')},
    )
    plateaus = find_plateaus(da, atol=sc.scalar(0.1, unit='m/s'), min_n_points=3)
    assert plateaus.sizes == {'plateau': 1}
    sc.testing.assert_identical(
        plateaus.coords['plateau'], sc.array(dims=['plateau'], values=[0], unit=None)
    )
    sc.testing.assert_identical(plateaus[0].value, da)


def test_find_plateaus_only_plateau_select_output_dim():
    da = sc.DataArray(
        sc.full(value=4.2, sizes={'x': 12}), coords={'x': sc.arange('x', 12)}
    )
    plateaus = find_plateaus(
        da, atol=sc.scalar(1e-7), min_n_points=3, plateau_dim='custom'
    )
    assert plateaus.sizes == {'custom': 1}
    sc.testing.assert_identical(
        plateaus.coords['custom'], sc.array(dims=['custom'], values=[0], unit=None)
    )
    sc.testing.assert_identical(plateaus[0].value, da)


def test_find_plateaus_no_plateau():
    da = sc.DataArray(
        sc.array(dims=['z'], values=[3, 6, 1, 2, 3, 2]), coords={'z': sc.arange('z', 6)}
    )
    plateaus = find_plateaus(da, atol=sc.scalar(0.1), min_n_points=2)
    assert plateaus.sizes == {'plateau': 0}
    sc.testing.assert_identical(
        plateaus.coords['plateau'],
        sc.array(dims=['plateau'], values=[], dtype=da.dtype, unit=None),
    )


def test_find_plateaus_plateaus_at_ends():
    """
    Plateau numbers:
    data ^
         |       1 1 1
         |     x
         | 0 0
         +------------>
                     t
    """
    da = sc.DataArray(
        sc.array(dims=['t'], values=[0, 0, 1, 2, 2, 2]), coords={'t': sc.arange('t', 6)}
    )
    plateaus = find_plateaus(da, atol=sc.scalar(0.1), min_n_points=2)
    assert plateaus.sizes == {'plateau': 2}
    sc.testing.assert_identical(
        plateaus.coords['plateau'], sc.array(dims=['plateau'], values=[0, 1], unit=None)
    )

    plateau0 = sc.DataArray(
        sc.array(dims=['t'], values=[0, 0]),
        coords={'t': sc.array(dims=['t'], values=[0, 1])},
    )
    sc.testing.assert_identical(plateaus[0].value, plateau0)
    plateau1 = sc.DataArray(
        sc.array(dims=['t'], values=[2, 2, 2]),
        coords={'t': sc.array(dims=['t'], values=[3, 4, 5])},
    )
    sc.testing.assert_identical(plateaus[1].value, plateau1)


def test_find_plateaus_slow_start_and_end():
    """
    Plateau numbers:
    data ^
         |         x
         |   1 1 1
         | x
         +------------>
                     t
    """
    da = sc.DataArray(
        sc.array(dims=['t'], values=[-3, -2, -2, -2, -1]),
        coords={'t': sc.arange('t', 5)},
    )
    plateaus = find_plateaus(da, atol=sc.scalar(0.1), min_n_points=2)
    assert plateaus.sizes == {'plateau': 1}
    sc.testing.assert_identical(
        plateaus.coords['plateau'], sc.array(dims=['plateau'], values=[0], unit=None)
    )

    plateau0 = sc.DataArray(
        sc.array(dims=['t'], values=[-2, -2, -2]),
        coords={'t': sc.array(dims=['t'], values=[1, 2, 3])},
    )
    sc.testing.assert_identical(plateaus[0].value, plateau0)


def test_find_plateaus_steep_start_and_end():
    """
    Plateau numbers:
    data ^
         | x
         |   x
         |     1 1
         |         x
         |           x
         +------------>
                     t
    """
    da = sc.DataArray(
        sc.array(dims=['t'], values=[1, 0, -1, -1, -1.5, -2]),
        coords={'t': sc.arange('t', 6)},
    )
    plateaus = find_plateaus(da, atol=sc.scalar(0.1), min_n_points=2)
    assert plateaus.sizes == {'plateau': 1}
    sc.testing.assert_identical(
        plateaus.coords['plateau'], sc.array(dims=['plateau'], values=[0], unit=None)
    )

    plateau0 = sc.DataArray(
        sc.array(dims=['t'], values=[-1.0, -1.0]),
        coords={'t': sc.array(dims=['t'], values=[2, 3])},
    )
    sc.testing.assert_identical(plateaus[0].value, plateau0)


def test_find_plateaus_peak():
    """
    Plateau numbers:
    data ^
         |       x
         |     x   x
         | 0 0
         +------------>
                     t
    """
    da = sc.DataArray(
        sc.array(dims=['t'], values=[10, 10, 20, 25, 20]),
        coords={'t': sc.arange('t', 5)},
    )
    plateaus = find_plateaus(da, atol=sc.scalar(0.1), min_n_points=2)
    assert plateaus.sizes == {'plateau': 1}
    sc.testing.assert_identical(
        plateaus.coords['plateau'], sc.array(dims=['plateau'], values=[0], unit=None)
    )

    plateau0 = sc.DataArray(
        sc.array(dims=['t'], values=[10, 10]),
        coords={'t': sc.array(dims=['t'], values=[0, 1])},
    )
    sc.testing.assert_identical(plateaus[0].value, plateau0)


def test_find_plateaus_adjacent_plateaus():
    """
    Plateau numbers:
    data ^
         |     1 1 1
         | 0 0
         +------------>
                     t
    """
    da = sc.DataArray(
        sc.array(dims=['t'], values=[0, 0, 1, 1, 1]), coords={'t': sc.arange('t', 5)}
    )
    plateaus = find_plateaus(da, atol=sc.scalar(0.1), min_n_points=2)
    assert plateaus.sizes == {'plateau': 2}
    sc.testing.assert_identical(
        plateaus.coords['plateau'], sc.array(dims=['plateau'], values=[0, 1], unit=None)
    )

    plateau0 = sc.DataArray(
        sc.array(dims=['t'], values=[0, 0]),
        coords={'t': sc.array(dims=['t'], values=[0, 1])},
    )
    sc.testing.assert_identical(plateaus[0].value, plateau0)
    plateau1 = sc.DataArray(
        sc.array(dims=['t'], values=[1, 1, 1]),
        coords={'t': sc.array(dims=['t'], values=[2, 3, 4])},
    )
    sc.testing.assert_identical(plateaus[1].value, plateau1)


def test_find_plateaus_adjacent_plateaus_select_long():
    """
    Plateau numbers:
    data ^
         |     1 1 1
         | 0 0
         +------------>
                     t
    """
    da = sc.DataArray(
        sc.array(dims=['t'], values=[0, 0, 1, 1, 1]), coords={'t': sc.arange('t', 5)}
    )
    plateaus = find_plateaus(da, atol=sc.scalar(0.1), min_n_points=3)
    assert plateaus.sizes == {'plateau': 1}
    sc.testing.assert_identical(
        plateaus.coords['plateau'], sc.array(dims=['plateau'], values=[0], unit=None)
    )

    plateau0 = sc.DataArray(
        sc.array(dims=['t'], values=[1, 1, 1]),
        coords={'t': sc.array(dims=['t'], values=[2, 3, 4])},
    )
    sc.testing.assert_identical(plateaus[0].value, plateau0)


def test_find_plateaus_adjacent_plateaus_select_none():
    """
    Plateau numbers:
    data ^
         |     1 1 1
         | 0 0
         +------------>
                     t
    """
    da = sc.DataArray(
        sc.array(dims=['t'], values=[0, 0, 1, 1, 1]), coords={'t': sc.arange('t', 5)}
    )
    plateaus = find_plateaus(da, atol=sc.scalar(0.1), min_n_points=4)
    assert plateaus.sizes == {'plateau': 0}
    sc.testing.assert_identical(
        plateaus.coords['plateau'],
        sc.array(dims=['plateau'], values=[], dtype=da.dtype, unit=None),
    )


def test_find_plateaus_slow_slope():
    da = sc.DataArray(
        sc.concat(
            [
                sc.linspace('s', 0.0, 1.0, 100, endpoint=False),
                sc.full(value=1.0, sizes={'s': 10}),
            ],
            dim='s',
        ),
        coords={'s': sc.arange('s', 110)},
    )
    plateaus = find_plateaus(da, atol=sc.scalar(1e-6), min_n_points=2)
    assert plateaus.sizes == {'plateau': 1}
    plateau0 = sc.DataArray(
        sc.full(value=1.0, sizes={'s': 10}), coords={'s': sc.arange('s', 100, 110)}
    )
    sc.testing.assert_identical(plateaus[0].value, plateau0)


def test_find_plateaus_slow_slope_false_positive_caught():
    da = sc.DataArray(sc.arange('s', 0.0, 1.0, 0.01), coords={'s': sc.arange('s', 100)})
    with pytest.raises(RuntimeError, match='tolerance'):
        find_plateaus(da, atol=sc.scalar(0.02), min_n_points=2)


def test_find_plateaus_plateaus_negative():
    """
    Plateau numbers:
                     t
         +------------>
         |           x
         |     1 1 1
         | 0 0
    data v
    """
    da = sc.DataArray(
        sc.array(dims=['t'], values=[-5, -5, -3, -3, -3, -2]),
        coords={'t': sc.arange('t', 6)},
    )
    plateaus = find_plateaus(da, atol=sc.scalar(0.1), min_n_points=2)
    assert plateaus.sizes == {'plateau': 2}
    sc.testing.assert_identical(
        plateaus.coords['plateau'], sc.array(dims=['plateau'], values=[0, 1], unit=None)
    )

    plateau0 = sc.DataArray(
        sc.array(dims=['t'], values=[-5, -5]),
        coords={'t': sc.array(dims=['t'], values=[0, 1])},
    )
    sc.testing.assert_identical(plateaus[0].value, plateau0)
    plateau1 = sc.DataArray(
        sc.array(dims=['t'], values=[-3, -3, -3]),
        coords={'t': sc.array(dims=['t'], values=[2, 3, 4])},
    )
    sc.testing.assert_identical(plateaus[1].value, plateau1)


def test_collapse_plateaus():
    """
    Plateau numbers:
    data ^
         |       1 1 1
         |     x       x
         | 0 0
         +--------------->
                        t
    """
    da = sc.DataArray(
        sc.array(dims=['t'], values=[0, 0, 1, 2, 2, 2, 1]),
        coords={
            't': sc.array(dims=['t'], values=[12, 13, 14, 15, 16, 17, 18], unit='s')
        },
    )
    plateaus = find_plateaus(da, atol=sc.scalar(0.1, unit='Hz'), min_n_points=2)
    collapsed = collapse_plateaus(plateaus, coord='t')

    t = sc.array(dims=['plateau', 't'], values=[[12, 14], [15, 18]], unit='s')
    t = t.transpose(collapsed.coords['t'].dims)
    expected = sc.DataArray(
        sc.array(dims=['plateau'], values=[0.0, 2.0]),
        coords={
            'plateau': sc.array(dims=['plateau'], values=[0, 1], unit=None),
            't': t,
        },
    )
    sc.testing.assert_identical(collapsed, expected)


def test_filter_in_phase_integers_positive():
    da = sc.DataArray(sc.array(dims=['t'], values=[7, 8, 8, 8, 3, 2, 4]))
    filtered = filter_in_phase(da, reference=sc.scalar(4), rtol=sc.scalar(0.1))
    expected = sc.DataArray(sc.array(dims=['t'], values=[8, 8, 8, 2, 4]))
    sc.testing.assert_identical(filtered, expected)


def test_filter_in_phase_integers_negative():
    da = sc.DataArray(sc.array(dims=['t'], values=[-7, -8, -8, -8, -3, -2, -4]))
    filtered = filter_in_phase(da, reference=sc.scalar(-4), rtol=sc.scalar(0.1))
    expected = sc.DataArray(sc.array(dims=['t'], values=[-8, -8, -8, -2, -4]))
    sc.testing.assert_identical(filtered, expected)


def test_filter_in_phase_floats():
    da = sc.DataArray(sc.array(dims=['t'], values=[0.23, 0.6, 1.2, 3.3, 2.4]))
    filtered = filter_in_phase(da, reference=sc.scalar(1.2), rtol=sc.scalar(0.1))
    expected = sc.DataArray(sc.array(dims=['t'], values=[0.6, 1.2, 2.4]))
    sc.testing.assert_identical(filtered, expected)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
import scipp as sc
from matplotlib.patches import Rectangle
from scipp.constants import h, m_n

from .._utils import as_float_type, elem_unit


def _tof_from_wavelength(
    *, wavelength: sc.Variable, Ltotal: sc.Variable
) -> sc.Variable:
    scale = (m_n / h).to(unit=sc.units.us / elem_unit(Ltotal) / elem_unit(wavelength))
    return as_float_type(Ltotal * scale, wavelength) * wavelength


class TimeDistanceDiagram:
    def __init__(self, ax, *, tmax, frame_rate=None):
        self._ax = ax
        self._time_unit = sc.Unit('ms')
        self._distance_unit = sc.Unit('m')
        frame_rate = 14.0 * sc.Unit('Hz') if frame_rate is None else frame_rate
        self._frame_length = (1.0 / frame_rate).to(unit=self._time_unit)
        self._tmax = tmax.to(unit=self._time_unit)
        self._ax.set_xlabel(f"time [{self._time_unit}]")
        self._ax.set_ylabel("distance [m]")

    @property
    def frame_length(self):
        return self._frame_length

    def to_time(self, time):
        return time.to(unit=self._time_unit)

    def to_distance(self, distance):
        return distance.to(unit=self._distance_unit)

    def annotate(self, text, *, xy, xytext, **kwargs):
        def to_mpl(point):
            x, y = point
            return self.to_time(x).value, self.to_distance(y).value

        self._ax.annotate(text, xy=to_mpl(xy), xytext=to_mpl(xytext), **kwargs)

    def add_source_pulse(self, pulse_length=None, ls='dotted'):
        pulse_length = 3.0 * sc.Unit('ms') if pulse_length is None else pulse_length
        t0 = 0.0
        t1 = self.to_time(pulse_length).value
        self._ax.text(
            t0,
            -2,
            f"Source pulse ({pulse_length.value} {pulse_length.unit})",
            ha="left",
            va="top",
            fontsize=6,
        )
        while t0 < self._tmax.value:
            rect = Rectangle((t0, 0), t1, -1, lw=1, fc='orange', ec='k')
            self._ax.add_patch(rect)
            self._ax.axvline(x=t0, ls=ls)
            t0 += self._frame_length.value

    def add_neutron(
        self,
        *,
        time_offset: sc.Variable,
        wavelength: sc.Variable,
        L: sc.Variable,
        label=None,
        color='black',
        ls='solid',
        lw=0.7,
    ):
        tof = self.to_time(_tof_from_wavelength(wavelength=wavelength, Ltotal=L))
        t0 = self.to_time(time_offset).value
        self._ax.plot(
            [t0, t0 + tof.value], [0, L.value], marker='', color=color, ls=ls, lw=lw
        )
        if label is not None:
            self._ax.text(tof.value, L.value, label, ha="center", va="bottom")

    def add_neutrons(
        self,
        *,
        lambda_min: sc.Variable,
        lambda_max: sc.Variable = None,
        Lmin: sc.Variable = 0.0 * sc.units.m,
        Lmax: sc.Variable,
        time_offset: sc.Variable,
        stride=1,
        frames=2,
    ):
        """
        Draw a wavelength band to depict propagation of neutrons. Neutrons are assumed
        to be emitted from a single point, i.e., no resolution effects are taken into
        account.

        Parameters
        ----------
        Lmin:
            Distance where neutrons are "emitted", such as the source pulse or
            a chopper. The default is at 0.0, i.e., the source position.
        Lmax:
            Distance where propagation of neutrons stops. This is typically set
            to the last detector (or after), but could be set to a chopper distance if
            a chopper extracts a smaller wavelength band.
        lambda_min:
            Minimum wavelength, defining fastest neutrons.
        lambda_max:
            Maximum wavelength, defining slowest neutrons. If lambda_max
            is None (the default) it is set such that there is no frame overlap at Lmax.
        time_offset:
            Offset time at which neutrons are emitted.
        frames:
            The number of frames that should be drawn.
        """
        tof_min = self.to_time(
            _tof_from_wavelength(wavelength=lambda_min, Ltotal=Lmax - Lmin)
        )
        if lambda_max is None:
            tof_max = tof_min + (stride - 1 + 0.95) * self._frame_length  # small 5% gap
        else:
            tof_max = self.to_time(
                _tof_from_wavelength(wavelength=lambda_max, Ltotal=Lmax - Lmin)
            )
        time_offset = self.to_time(time_offset)
        t0 = time_offset
        tmin = t0 + tof_min
        tmax = t0 + tof_max
        t = sc.concat([t0, t0, tmax, tmin], 't')
        L = sc.concat([Lmin, Lmin, Lmax, Lmax], 'L')
        for _ in range(frames):
            self._ax.fill(t.values, L.values, alpha=0.3)
            t += stride * self._frame_length

    def add_detector(self, *, distance, name='detector'):
        # TODO This could accept a list of positions and plot a rectangle from min to
        # max detector distance
        self._ax.plot(
            [0, self._tmax.max().value],
            [distance.value, distance.value],
            lw=3,
            color='grey',
        )
        self._ax.text(0.0, distance.value, name, va="bottom", ha="left")

    def add_sample(self, *, distance):
        self._ax.plot(
            [0, self._tmax.max().value],
            [distance.value, distance.value],
            lw=3,
            color='green',
        )
        self._ax.text(0.0, distance.value, 'sample', va="bottom", ha="left")



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
"""
A fake time-of-flight neutron beamline for documentation and testing.

This provides data in a structure as typically provided in a NeXus file, including:

- Detector event data including event_time_offset and event_time_zero
- Monitor event data including event_time_offset and event_time_zero
- Chopper timestamps
"""

from __future__ import annotations

from collections.abc import Callable

import scipp as sc
from numpy import random

from . import chopper_cascade


class FakeSource:
    def __init__(
        self,
        frequency: sc.Variable,
        run_length: sc.Variable,
        events_per_pulse: int = 10000,
    ):
        """
        Return a fake source.

        Parameters
        ----------
        frequency:
            Frequency of the source.
        run_length:
            Run length of the source.
        events_per_pulse:
            Number of events per pulse.
        """
        self.frequency = frequency
        self.t0 = self._make_t0(frequency=frequency, run_length=run_length)
        self.events_per_pulse = events_per_pulse
        self.rng = random.default_rng(seed=0)

    @property
    def pulse_period(self) -> sc.Variable:
        return sc.scalar(1.0) / self.frequency

    @property
    def number_of_pulses(self) -> int:
        return len(self.t0)

    def _make_t0(self, frequency: sc.Variable, run_length: sc.Variable) -> sc.Variable:
        start = sc.datetime('2019-12-25T06:00:00.0', unit='ns')
        npulse = (run_length * frequency).to(unit='').value
        pulses = sc.arange(dim='pulse', start=0, stop=npulse, dtype='int64')
        return start + (pulses * (1.0 / frequency)).to(dtype='int64', unit='ns')


class FakePulse:
    """
    Simplified model of a pulse.

    Currently this is simply a time and wavelength interval. The plan is to also model
    a tail in the future, e.g., by overlaying multiple pulses.
    """

    def __init__(
        self,
        time_min: sc.Variable,
        time_max: sc.Variable,
        wavelength_min: sc.Variable,
        wavelength_max: sc.Variable,
    ):
        self.time_min = time_min
        self.time_max = time_max
        self.wavelength_min = wavelength_min
        self.wavelength_max = wavelength_max
        self._center = sc.scalar(0.5) * (time_min + time_max)
        self._frames = chopper_cascade.FrameSequence.from_source_pulse(
            time_min=time_min,
            time_max=time_max,
            wavelength_min=wavelength_min,
            wavelength_max=wavelength_max,
        )

    @property
    def center(self) -> sc.Variable:
        return self._center

    def chop(
        self, choppers: list[chopper_cascade.Chopper]
    ) -> chopper_cascade.FrameSequence:
        return self._frames.chop(choppers)


class FakeBeamline:
    def __init__(
        self,
        source: FakeSource,
        pulse: FakePulse,
        choppers: dict[str, chopper_cascade.Chopper],
        monitors: dict[str, sc.Variable],
        detectors: dict[str, sc.Variable],
        time_of_flight_origin: str | None = None,
    ):
        """
        Return a fake beamline.

        Parameters
        ----------
        source:
            Fake source.
        pulse:
            Fake pulse.
        choppers:
            Choppers.
        monitors:
            Distances of monitors.
        detectors:
            Distances of detectors.
        time_of_flight_origin:
            Name of the chopper to use as time-of-flight origin. If None, use the
            source pulse. The center time of the slit opening of the chopper (or the
            source pulse) is used as time-of-flight origin.
        """
        self._frames = pulse.chop(choppers.values())
        self._pulse = pulse
        self._choppers = choppers
        self._source = source
        self._monitors = {
            key: self._frames[distance] for key, distance in monitors.items()
        }
        self.detectors = {
            key: self._frames[distance] for key, distance in detectors.items()
        }
        self._time_of_flight_origin = time_of_flight_origin

    def get_monitor(self, name: str) -> sc.DataGroup:
        frame = self._monitors[name]
        return self._fake_monitor(frame)

    def _split_size(self, size, N):
        base, remainder = divmod(size, N)
        sizes = [base + 1 if i < remainder else base for i in range(N)]
        return sizes

    def _fake_monitor(
        self, frame: chopper_cascade.Frame
    ) -> tuple[sc.DataArray, sc.DataArray]:
        bounds = frame.bounds()['time']
        subbounds = frame.subbounds()['time']
        subframes = subbounds.sizes['subframe']

        sizes = sc.array(
            dims=['pulse'],
            values=self._source.rng.integers(
                0, self._source.events_per_pulse, size=self._source.number_of_pulses
            ),
            unit=None,
        )
        event_index = sc.cumsum(sizes, dim='pulse', mode='exclusive')
        size = sizes.sum().value
        subsizes = self._split_size(size, subframes)
        subframe_times = [
            sc.array(
                dims=['event'],
                values=self._source.rng.uniform(
                    subbounds['subframe', i][0].value,
                    subbounds['subframe', i][-1].value,
                    size=subsizes[i],
                ),
                unit=bounds.unit,
            )
            for i in range(subframes)
        ]

        # Offset from pulse that created the monitor event
        time_offset = sc.concat(subframe_times, 'event')
        # Ensure all pulses have events from all subframes
        self._source.rng.shuffle(time_offset.values)
        event_time_offset = time_offset % self._source.pulse_period
        time_zero_offset = time_offset - event_time_offset
        event_time_zero = self._source.t0
        wrapped_events = sc.DataArray(
            sc.ones(sizes=event_time_offset.sizes, unit='counts'),
            coords={'event_time_offset': event_time_offset},
        )
        unwrapped_events = sc.DataArray(
            sc.ones(sizes=time_offset.sizes, unit='counts'),
            coords={
                'time_offset': time_offset,
                'time_zero_offset': time_zero_offset.to(dtype='int64', unit='ns'),
            },
        )
        wrapped = sc.DataArray(
            data=sc.bins(begin=event_index, dim='event', data=wrapped_events),
            coords={'event_time_zero': event_time_zero},
        )
        unwrapped = sc.DataArray(
            data=sc.bins(begin=event_index, dim='event', data=unwrapped_events),
            coords={'event_time_zero': event_time_zero},
        )
        if self._time_of_flight_origin is None:
            offset_to_tof = self._pulse.center
        else:
            source_chopper = self._choppers[self._time_of_flight_origin]
            if len(source_chopper.time_open) != 1:
                raise NotImplementedError(
                    "Using a chopper with multiple openings as "
                    "source chopper is not implemented yet."
                )
            offset_to_tof = 0.5 * (
                source_chopper.time_open[0] + source_chopper.time_close[0]
            )
        unwrapped = unwrapped.transform_coords(
            tof=lambda time_offset: time_offset - offset_to_tof.to(unit='s'),
            time_zero=lambda event_time_zero, time_zero_offset: event_time_zero
            + time_zero_offset.to(dtype='int64', unit='ns')
            + offset_to_tof.to(dtype='int64', unit='ns'),
        )
        if self._time_of_flight_origin is None:
            unwrapped.coords['Ltotal'] = frame.distance
        else:
            unwrapped.coords['Ltotal'] = frame.distance - source_chopper.distance
        return wrapped, unwrapped


class FakeBeamlineEss:
    def __init__(
        self,
        choppers: dict[str, chopper_cascade.Chopper],
        monitors: dict[str, sc.Variable],
        run_length: sc.Variable,
        events_per_pulse: int = 200000,
        source: Callable | None = None,
        # time_of_flight_origin: str | None = None,
    ):
        import math

        import tof as tof_pkg
        from tof.facilities.ess_pulse import pulse

        self.frequency = pulse.frequency
        self.npulses = math.ceil((run_length * self.frequency).to(unit='').value)
        self.events_per_pulse = events_per_pulse

        # Create a source
        if source is None:
            self.source = tof_pkg.Source(
                facility='ess', neutrons=self.events_per_pulse, pulses=self.npulses
            )
        else:
            self.source = source(pulses=self.npulses)

        # Convert the choppers to tof.Chopper
        angular_speed = sc.constants.pi * (2.0 * sc.units.rad) * self.frequency

        self.choppers = [
            tof_pkg.Chopper(
                frequency=pulse.frequency,
                open=ch.time_open * angular_speed,
                close=ch.time_close * angular_speed,
                phase=sc.scalar(0.0, unit='rad'),
                distance=ch.distance,
                name=name,
            )
            for name, ch in choppers.items()
        ]

        # Add detectors
        self.monitors = [
            tof_pkg.Detector(distance=distance, name=key)
            for key, distance in monitors.items()
        ]

        #  Propagate the neutrons
        self.model = tof_pkg.Model(
            source=self.source, choppers=self.choppers, detectors=self.monitors
        )
        self.model_result = self.model.run()

    def get_monitor(self, name: str) -> sc.DataGroup:
        # Create some fake pulse time zero
        start = sc.datetime("2024-01-01T12:00:00.000000")
        period = sc.reciprocal(self.frequency)

        # Format the data in a way that resembles data loaded from NeXus
        event_data = self.model_result.detectors[name].data.flatten(to='event')
        dt = period.to(unit='us')
        event_data.coords['event_time_zero'] = (
            dt * (event_data.coords['toa'] // dt)
        ).to(dtype=int) + start
        event_data.coords['event_time_offset'] = (
            event_data.coords.pop('tof').to(unit='s') % period
        )
        del event_data.coords['speed']
        del event_data.coords['time']
        del event_data.coords['wavelength']

        # Select only the neutrons that make it to the detector
        out = (
            event_data[~event_data.masks['blocked_by_others']]
            .group('event_time_zero')
            .rename_dims(event_time_zero='pulse')
        )
        return out


wfm1 = chopper_cascade.Chopper(
    distance=sc.scalar(6.6, unit='m'),
    time_open=sc.array(
        dims=('cutout',),
        values=[
            -0.000396,
            0.001286,
            0.005786,
            0.008039,
            0.010133,
            0.012080,
            0.013889,
            0.015571,
        ],
        unit='s',
    ),
    time_close=sc.array(
        dims=('cutout',),
        values=[
            0.000654,
            0.002464,
            0.006222,
            0.008646,
            0.010899,
            0.012993,
            0.014939,
            0.016750,
        ],
        unit='s',
    ),
)
wfm2 = chopper_cascade.Chopper(
    distance=sc.scalar(7.1, unit='m'),
    time_open=sc.array(
        dims=('cutout',),
        values=[
            0.000654,
            0.002451,
            0.006222,
            0.008645,
            0.010898,
            0.012993,
            0.014940,
            0.016737,
        ],
        unit='s',
    ),
    time_close=sc.array(
        dims=('cutout',),
        values=[
            0.001567,
            0.003641,
            0.006658,
            0.009252,
            0.011664,
            0.013759,
            0.015853,
            0.017927,
        ],
        unit='s',
    ),
)
# psc1 and psc2 are defined as a single slit of wfm1 and wfm2, respectively.
psc1 = chopper_cascade.Chopper(
    distance=sc.scalar(6.6, unit='m'),
    time_open=sc.array(dims=('cutout',), values=[0.010133], unit='s'),
    time_close=sc.array(dims=('cutout',), values=[0.010899], unit='s'),
)
psc2 = chopper_cascade.Chopper(
    distance=sc.scalar(7.1, unit='m'),
    time_open=sc.array(dims=('cutout',), values=[0.010898], unit='s'),
    time_close=sc.array(dims=('cutout',), values=[0.011664], unit='s'),
)
frame_overlap_1 = chopper_cascade.Chopper(
    distance=sc.scalar(8.8, unit='m'),
    time_open=sc.array(
        dims=('cutout',),
        values=[
            -0.000139,
            0.002460,
            0.006796,
            0.010020,
            0.012733,
            0.015263,
            0.017718,
            0.020317,
        ],
        unit='s',
    ),
    time_close=sc.array(
        dims=('cutout',),
        values=[
            0.000640,
            0.003671,
            0.007817,
            0.011171,
            0.013814,
            0.016146,
            0.018497,
            0.021528,
        ],
        unit='s',
    ),
)
frame_overlap_2 = chopper_cascade.Chopper(
    distance=sc.scalar(15.9, unit='m'),
    time_open=sc.array(
        dims=('cutout',),
        values=[
            -0.000306,
            0.010939,
            0.016495,
            0.021733,
            0.026416,
            0.030880,
            0.035409,
        ],
        unit='s',
    ),
    time_close=sc.array(
        dims=('cutout',),
        values=[
            0.002582,
            0.014570,
            0.020072,
            0.024730,
            0.029082,
            0.033316,
            0.038297,
        ],
        unit='s',
    ),
)
pulse_overlap = chopper_cascade.Chopper(
    distance=sc.scalar(22.0, unit='m'),
    time_open=sc.array(dims=('cutout',), values=[-0.130952, 0.011905], unit='s'),
    time_close=sc.array(dims=('cutout',), values=[-0.087302, 0.055556], unit='s'),
)

wfm_choppers = sc.DataGroup(
    wfm1=wfm1,
    wfm2=wfm2,
    frame_overlap_1=frame_overlap_1,
    frame_overlap_2=frame_overlap_2,
    pulse_overlap=pulse_overlap,
)
psc_choppers = sc.DataGroup(
    psc1=psc1,
    psc2=psc2,
    frame_overlap_1=frame_overlap_1,
    frame_overlap_2=frame_overlap_2,
    pulse_overlap=pulse_overlap,
)

ess_time_min = sc.scalar(0.0, unit='ms')
ess_time_max = sc.scalar(3.0, unit='ms')
ess_wavelength_min = sc.scalar(0.0, unit='angstrom')
ess_wavelength_max = sc.scalar(10.0, unit='angstrom')

wfm_frames = chopper_cascade.FrameSequence.from_source_pulse(
    time_min=ess_time_min,
    time_max=ess_time_max,
    wavelength_min=ess_wavelength_min,
    wavelength_max=ess_wavelength_max,
)
wfm_frames = wfm_frames.chop(wfm_choppers.values())

psc_frames = chopper_cascade.FrameSequence.from_source_pulse(
    time_min=ess_time_min,
    time_max=ess_time_max,
    wavelength_min=ess_wavelength_min,
    wavelength_max=ess_wavelength_max,
)
psc_frames = psc_frames.chop(psc_choppers.values())



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Jan-Lukas Wynen

import scipp as sc

from .conversion import graph


def _derived_coord(
    da: sc.DataArray | sc.Dataset, name: str, scatter: bool = True
) -> sc.Variable:
    tmp = da.transform_coords(
        name,
        graph=graph.beamline.beamline(scatter=scatter),
        rename_dims=False,
        keep_aliases=False,
        keep_inputs=False,
        keep_intermediate=False,
    )
    return tmp.coords[name]


def position(da: sc.DataArray | sc.Dataset) -> sc.Variable:
    """Extract the detector pixel positions from a data array or dataset.

    Parameters
    ----------
    da:
        Get or compute the positions from coords and attrs of this.

    Returns
    -------
    :
        The detector pixel positions.
    """
    return _derived_coord(da, 'position')


def source_position(da: sc.DataArray | sc.Dataset) -> sc.Variable:
    """Extract the position of the neutron source from a data array or dataset.

    Parameters
    ----------
    da:
        Get or compute the source position from coords and attrs of this.

    Returns
    -------
    :
        The source position.
    """
    return _derived_coord(da, 'source_position')


def sample_position(da: sc.DataArray | sc.Dataset) -> sc.Variable:
    """Extract the position of the sample from a data array or dataset.

    Parameters
    ----------
    da:
        Get or compute the sample position from coords and attrs of this.

    Returns
    -------
    :
        The sample position.
    """
    return _derived_coord(da, 'sample_position')


def incident_beam(da: sc.DataArray | sc.Dataset) -> sc.Variable:
    """Extract the incident beam vector from a data array or dataset.

    This is the direction and length of the primary flight path,
    i.e. from source to sample.

    Parameters
    ----------
    da:
        Get or compute the incident beam from coords and attrs of this.

    Returns
    -------
    :
        The incident beam.
    """
    return _derived_coord(da, 'incident_beam')


def scattered_beam(da: sc.DataArray | sc.Dataset) -> sc.Variable:
    """Extract the scattered beam vector from a data array or dataset.

    This is the direction and length of the secondary flight path,
    i.e. from sample to detector.

    Parameters
    ----------
    da:
        Get or compute the scattered beam from coords and attrs of this.

    Returns
    -------
    :
        The scattered beam.
    """
    return _derived_coord(da, 'scattered_beam')


def Ltotal(da: sc.DataArray | sc.Dataset, scatter: bool) -> sc.Variable:
    """Extract the length of the total flight path from a data array or dataset.

    Parameters
    ----------
    da:
        Get or compute the total flight path length from coords and attrs of this.
    scatter:
        If ``True``, assume a beam that scattered off a sample and return the
        sum of L1 and L2.
        If ``False``, return the straight distance between source and detector.

    Returns
    -------
    :
        The length of the total flight path.
    """
    return _derived_coord(da, 'Ltotal', scatter=scatter)


def L1(da: sc.DataArray | sc.Dataset) -> sc.Variable:
    """Extract the length of the primary flight path from a data array or dataset.

    This is the distance between neutron source and sample.

    Parameters
    ----------
    da:
        Get or compute the primary flight path length from coords and attrs of this.

    Returns
    -------
    :
        The length of the primary flight path.
    """
    return _derived_coord(da, 'L1')


def L2(da: sc.DataArray | sc.Dataset) -> sc.Variable:
    """Extract the length of the secondary flight path from a data array or dataset.

    This is the distance between sample and detector.

    Parameters
    ----------
    da:
        Get or compute the secondary flight path length from coords and attrs of this.

    Returns
    -------
    :
        The length of the secondary flight path.
    """
    return _derived_coord(da, 'L2')


def two_theta(da: sc.DataArray | sc.Dataset) -> sc.Variable:
    """Extract the scattering angle from a data array or dataset.

    The angle is defined as in Bragg's law.
    See the beamline `documentation <../modules/scippneutron.conversion.beamline.rst>`_
    for a full definition.

    Parameters
    ----------
    da:
        Get or compute the scattering angle from coords and attrs of this.

    Returns
    -------
    :
        Twice the scattering angle.
    """
    return _derived_coord(da, 'two_theta')
